"""
This contains the Content implementation for GitLab.
"""
from typing import Optional
from typing import Union
from urllib.parse import quote_plus

from IGitt.GitLab import GitLabMixin
from IGitt.GitLab import GitLabOAuthToken
from IGitt.GitLab import GitLabPrivateToken
from IGitt.Interfaces import delete
from IGitt.Interfaces import get
from IGitt.Interfaces import put
from IGitt.Interfaces.Content import Content


class GitLabContent(GitLabMixin, Content):
    """
    This class represents a content on GitHub
    """
    def __init__(self,  token: Union[GitLabOAuthToken, GitLabPrivateToken],
                 repository: str, path: str):
        self._token = token
        self._repository = repository
        self._url = ('/projects/' + quote_plus(repository) +
                     '/repository/files/' + path)

    def get_content(self, ref='master'):
        data = {
            'path': self._url,
            'ref': ref
        }
        self.data = get(token=self._token, url=self.url, params=data)

    def delete(self, message: str, branch: Optional[str]=None):
        """
        Deletes content

        :param message: The commit message for the deletion commit.
        :param branch:  The branch to delete this content from. Defaults to
                        `master`.
        """
        if branch is None:
            branch = 'master'

        data = {
            'file_path': self._url,
            'branch': branch,
            'commit_message': message
        }
        delete(self._token, url=self.url, params=data)

    def update(self, message: str, content: str, branch: Optional[str]=None):

        if branch is None:
            branch = 'master'

        data = {
            'file_path': self._url,
            'commit_message': message,
            'branch': branch,
            'content' : content
        }
        put(token=self._token, url=self.url, data=data)



"""
This contains the Issue implementation for GitHub.
"""
from datetime import datetime
from typing import Optional
from typing import Set
import re

import requests

from IGitt.GitHub import GitHubMixin, GH_INSTANCE_URL
from IGitt.GitHub import GitHubToken
from IGitt.GitHub.GitHubComment import GitHubComment
from IGitt.GitHub.GitHubReaction import GitHubReaction
from IGitt.GitHub.GitHubReaction import PREVIEW_HEADER
from IGitt.GitHub.GitHubUser import GitHubUser
from IGitt.Interfaces.Comment import CommentType
from IGitt.Interfaces.Issue import Issue
from IGitt.Interfaces import get, patch, post, delete
from IGitt.Interfaces import IssueStates


CLOSED_BY_PATTERN = re.compile('closed this(?:\n| )+in(?:\n| )+<a href=\"/(.+)/'
                               'pull/([0-9]+)\">#(?:[0-9]+)</a>')


class GitHubIssue(GitHubMixin, Issue):
    """
    This class represents an issue on GitHub.
    """

    def __init__(self, token: GitHubToken, repository: str, number: int):
        """
        Creates a new GitHubIssue with the given credentials.

        >>> from os import environ
        >>> issue = GitHubIssue(environ['GITHUB_TEST_TOKEN'],
        ...                     'gitmate-test-user/repo_that_doesnt_exist', 1)
        ... # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
         ...
        RuntimeError: ({'message': 'Not Found', ...}, 404)

        :param token: A GitHubToken object.
        :param repository: The full name of the repository,
                           e.g. ``sils/something``.
        :param number: The issue number.
        :raises RuntimeError: If something goes wrong (network, auth, ...)
        """
        self._token = token
        self._repository = repository
        self._number = number
        self._url = '/repos/'+repository+'/issues/'+str(number)

    @property
    def repository(self):
        """
        Returns the GitHub repository this issue is linked with as a
        GitHubRepository instance.
        """
        from IGitt.GitHub.GitHubRepository import GitHubRepository
        return GitHubRepository(self._token, self._repository)

    @property
    def title(self):
        """
        Retrieves the title of the issue.

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> issue.title
        'test issue'

        You can simply set it using the property setter:

        >>> issue.title = 'dont panic'
        >>> issue.title
        'dont panic'

        >>> issue.title = 'test issue'

        :return: The title of the issue - as string.
        """
        return self.data['title']

    @title.setter
    def title(self, new_title):
        """
        Sets the title of the issue.

        :param new_title: The new title.
        """
        self.data = patch(self._token, self.url, {'title': new_title})

    @property
    def number(self) -> int:
        """
        Returns the issue "number" or id.

        >>> from os import environ
        >>> issue = GitHubIssue(environ['GITHUB_TEST_TOKEN'],
        ...                     'gitmate-test-user/test', 1)
        >>> issue.number
        1

        :return: The number of the issue.
        """
        return self._number

    @property
    def assignees(self) -> Set[GitHubUser]:
        """
        Retrieves the assignee of the issue:

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> {a.username for a in issue.assignees}
        {'gitmate-test-user'}

        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 2)
        >>> issue.assignees  # Returns empty set, unassigned
        {}

        :return: A set containing the usernames of assignees.
        """
        return set(
            GitHubUser.from_data(user, self._token, user['login'])
            for user in self.data['assignees']
        )

    @assignees.setter
    def assignees(self, value: Set[GitHubUser]):
        """
        Setter for ssignees.
        """
        if value - self.assignees:
            self.assign(*(value - self.assignees))

        if self.assignees - value:
            self.unassign(*(self.assignees - value))

    def assign(self, *users: Set[GitHubUser]):
        """
        Adds the user as one of the assignees of the issue.
        :param username: Username of the user to be added as an assignee.
        """
        url = self.url + '/assignees'
        self.data = post(self._token, url,
                         {'assignees': [user.username for user in users]})

    def unassign(self, *users: Set[GitHubUser]):
        """
        Removes the user from the assignees of the issue.
        :param users: Username of the user to be unassigned.
        """
        url = self.url + '/assignees'
        delete(self._token, url,
               {'assignees': [user.username for user in users]})
        self.data = get(self._token, self.url)

    @property
    def available_assignees(self) -> Set[GitHubUser]:
        """
        Retrieves a set of available assignees for the issue.

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> {a.username for a in issue.available_assignees}
        {'gitmate-test-user'}

        :return: A set of GitHubUsers.
        """
        return {
            GitHubUser.from_data(user, self._token, user['login'])
            for user in get(
                self._token,
                self.absolute_url('/repos/' + self._repository + '/assignees'))
        }

    @property
    def description(self):
        r"""
        Retrieves the main description of the issue:

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> issue.description
        'A nice description!\n'

        :return: A string containing the main description of the issue.
        """
        return self.data['body'] if self.data['body'] else ''

    @description.setter
    def description(self, new_description):
        """
        Sets the description of the issue.

        :param new_description: The new description.
        """
        self.data = patch(self._token, self.url, {'body': new_description})

    @property
    def author(self) -> GitHubUser:
        """
        Retrieves the author of the issue.

        :return: A GitHubUser object.
        """
        return GitHubUser.from_data(self.data['user'],
                                    self._token,
                                    self.data['user']['login'])

    def add_comment(self, body):
        """
        Adds a comment to the issue:

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 3)
        >>> comment = issue.add_comment("Doh!")

        You can use the comment right after:

        >>> comment.body
        'Doh!'
        >>> comment.delete()

        The comment will be created by the user authenticated via the oauth
        token.

        :param body: The body of the new comment to create.
        :return: The newly created comment.
        """
        result = post(self._token, self.url + '/comments', {'body': body})

        return GitHubComment.from_data(result, self._token, self._repository,
                                       CommentType.ISSUE, result['id'])

    @property
    def comments(self):
        r"""
        Retrieves comments from the issue.

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 9)
        >>> comments = issue.comments

        Now we can e.g. access the last comment:

        >>> comments[-1].body
        'Do not comment here.\n'

        :return: A list of Comment objects.
        """
        return [GitHubComment.from_data(result, self._token, self._repository,
                                        CommentType.ISSUE, result['id'])
                for result in get(self._token, self.url + '/comments')]

    @property
    def labels(self):
        """
        Retrieves all labels associated with this bug.

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> issue.labels
        set()

        Use the property setter to set the labels:

        >>> issue.labels = {'a', 'b', 'c'}
        >>> sorted(issue.labels)
        ['a', 'b', 'c']

        Use the empty set intuitively to clear all labels:

        >>> issue.labels = set()

        :return: A list of label captions (str).
        """
        return set(label['name'] for label in self.data['labels'])

    @labels.setter
    def labels(self, value: Set[str]):
        """
        Sets the labels to the given set of labels.

        :param value: A set of label texts.
        """
        # Only if self.data is populated we actually save a request here
        if 'labels' in self.data and value == self.labels:
            return  # No need to patch

        self.data = patch(self._token, self.url, {'labels': list(value)})

    @property
    def available_labels(self):
        """
        Retrieves a set of captions that are available for labelling bugs.

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> sorted(issue.available_labels)
        ['a', 'b', 'c']

        :return: A set of label captions (str).
        """
        return {label['name'] for label in get(
            self._token,
            self.absolute_url('/repos/' + self._repository + '/labels'))}

    @property
    def created(self) -> datetime:
        """
        Retrieves a timestamp on when the issue was created.

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> issue.created
        datetime.datetime(2016, 1, 13, 7, 56, 23)
        """
        return datetime.strptime(self.data['created_at'],
                                 '%Y-%m-%dT%H:%M:%SZ')

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the issue was updated the last time.

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 9)
        >>> issue.updated
        datetime.datetime(2016, 10, 9, 11, 27, 11)
        """
        return datetime.strptime(self.data['updated_at'],
                                 '%Y-%m-%dT%H:%M:%SZ')

    def close(self):
        """
        Closes the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        self.data = patch(self._token, self.url, {'state': 'closed'})

    def reopen(self):
        """
        Reopens the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        self.data = patch(self._token, self.url, {'state': 'open'})

    def delete(self):
        """
        Should delete the issue, but GitHub doesn't allow it yet.

        Reference: https://github.com/isaacs/github/issues/253
        """
        raise NotImplementedError("GitHub doesn't allow deleting issues.")

    @property
    def state(self) -> str:
        """
        Get's the state of the issue.

        >>> from os import environ
        >>> issue = GitHubIssue(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 10)
        >>> issue.state
        <IssueStates.OPEN: 'open'>
        >>> str(issue.state)
        'open'

        So if we close it:

        >>> issue.close()
        >>> issue.state
        <IssueStates.CLOSED: 'closed'>
        >>> str(issue.state)
        'closed'

        And reopen it:

        >>> issue.reopen()
        >>> issue.state
        <IssueStates.OPEN: 'open'>


        :return:    Either <IssueStates.OPEN: 'open'> or
                    <IssueStates.CLOSED: 'closed'>.
        """
        return IssueStates[self.data['state'].upper()]

    @property
    def reactions(self) -> Set[GitHubReaction]:
        """
        Retrieves the reactions / award emojis applied on the issue.
        """
        url = self.url + '/reactions'
        reactions = get(self._token, url, headers=PREVIEW_HEADER)
        return {GitHubReaction.from_data(r, self._token, self, r['id'])
                for r in reactions}

    @staticmethod
    def create(token: str, repository: str,
               title: str, body: str='', issue_type: Optional[str]=None):
        """
        Create a new issue with given title and body.

        >>> from os import environ
        >>> issue = GitHubIssue.create(
        ...     GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...     'gitmate-test-user/test',
        ...     'test issue title',
        ...     'sample description'
        ... )
        >>> issue.state
        <IssueStates.OPEN: 'open'>
        >>> str(issue.state)
        'open'

        Let's delete the newly created one, because it's useless.

        >>> issue.delete()
        ... # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
         ...
        NotImplementedError

        Because GitHub is stupid, they don't allow deleting issues. So, let's
        atleast close this for now.

        >>> issue.close()

        :return: GitHubIssue object of the newly created issue.
        """

        post_url = '/repos/' + repository + '/issues'
        data = {
            'title': title,
            'body': body,
        }

        resp = post(token, GitHubIssue.absolute_url(post_url), data)
        issue_number = resp['number']

        return GitHubIssue.from_data(resp, token, repository, issue_number)

    @property
    def mrs_closed_by(self):
        """
        Returns the merge requests that close this issue.
        """
        from IGitt.GitHub.GitHubMergeRequest import GitHubMergeRequest

        r = requests.get(GH_INSTANCE_URL + self._url.replace('/repos', ''))

        matches = CLOSED_BY_PATTERN.findall(r.text)

        return {GitHubMergeRequest(self._token, repo_name, int(number))
                for repo_name, number in matches}

    @property
    def milestone(self):
        """
        Retrieves the milestone.
        """
        from IGitt.GitHub.GitHubMilestone import GitHubMilestone
        return GitHubMilestone.from_data(
            self.data['milestone'], self._token, self._repository,
            self.data['milestone']['number']
        ) if self.data['milestone'] else None

    @milestone.setter
    def milestone(self, new_milestone):
        """
        Setter for the Milestone.
        Delete the Milestone with passing a 'None'
        """
        self.data = patch(
            self._token, self.url,
            {'milestone': new_milestone.number if new_milestone else ''})



"""
This module contains the Installation abstraction class which provides
properties and actions related to integration installations on supported
providers.
"""
from typing import Set
from typing import List

from IGitt.Interfaces import IGittObject
from IGitt.Interfaces.Repository import Repository


class Installation(IGittObject):
    """
    Represents an application Installation on supported providers.
    """

    @property
    def identifier(self) -> int:
        """
        Returns the identifier of the installation.
        """
        raise NotImplementedError

    @property
    def app_id(self) -> int:
        """
        The application which was installed.
        """
        raise NotImplementedError

    @property
    def webhooks(self) -> List[str]:
        """
        Returns the list of webhooks registered by installing the application.
        """
        raise NotImplementedError

    @property
    def permissions(self):
        """
        The permissions granted to this installation.
        """
        raise NotImplementedError

    @property
    def repositories(self) -> Set[Repository]:
        """
        Returns the set of repositories this installation has access to.
        """
        raise NotImplementedError



"""
This module contains the helpers for writing testcases for IGitt.
"""
from abc import ABCMeta
from inspect import getfile
import os
from os.path import dirname
from os.path import join
from unittest import TestCase
import re

from vcr import VCR
import pytest


FILTER_QUERY_PARAMS = ['access_token', 'private_token']
FILTER_PARAMS_REGEX = re.compile(r'(\??)((?:{})=\w+&?)'.format(
    '|'.join(FILTER_QUERY_PARAMS)))


# Random private key
PRIVATE_KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIICXAIBAAKBgQCZ3CJJ+7vflLg6XPG59kWjfALWJlRtVj1YWmr95dR5a8FRsnwF
Lqgh2vLT2lk2Q6PoXX1C5blsM/xrf0Hc/fw2aOE/RvDCgHG83IBcFv8YOEndgTJp
qBmEK0xIbxhdUtku4xpiZddzyt8KoRg7uHM5P06TX5qJjK5JKL/B4RKnNwIDAQAB
AoGAIxwLqwxJu+RpAdBxzLi4/WxwDUQj4etbBk1jutp2WNrQ+36aNGiIL2mSHevm
ja5zubOTwO9BF8LpJ/KbKf2/TqPlKH4QZibHqoPSAHa6n5RHQBn4f6EauG4IaJfC
nLA5ZyP3T29uCpByM745LUOpteEj/fQCxjYE2y+MsCzdw8ECQQDhMK7psIE7INbN
cUjikA6F0h48ZpDBYPYy+EetwLc45i/fWLd920DxkuQ6zu/qWQWY/x+ZNh/WoYfr
GqVpgszhAkEArukZHcd7z7OCso+2SvWmEmrMprfcOpw/1JsIMt2W9GhnBHe+wBvv
f4jY5iRl/5dKfNXppkz1G1GiXrx7ZOMfFwJADWd1cemUt61Lu+zbVskWZDbOn+/G
/AvGe+A1fA01mshw3w2L1oz/f6GrvihlNYDZCXNeMSN8n6z7xy3N3MrxYQJAHWcw
ArKLHLJXkT7ZbSZ4YXY0qv4TdoLXtBzPtwVLIBEA6F5c4ZyQmUbe92k9AEdljTDE
k2EyfwItInHa6G3JxwJBAIYfzh6YmG/hv0rTgBgKXX0NL+BS/ppcy85w4zg08e/f
Edf3/k7WkC18G802pLT5+xf8snh+Oti95jfoHU1j4ek=
-----END RSA PRIVATE KEY-----
"""

PUBLIC_KEY = """
-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQCZ3CJJ+7vflLg6XPG59kWjfALW
JlRtVj1YWmr95dR5a8FRsnwFLqgh2vLT2lk2Q6PoXX1C5blsM/xrf0Hc/fw2aOE/
RvDCgHG83IBcFv8YOEndgTJpqBmEK0xIbxhdUtku4xpiZddzyt8KoRg7uHM5P06T
X5qJjK5JKL/B4RKnNwIDAQAB
-----END PUBLIC KEY-----
"""

# set environment variables default
ENV_DEFAULTS = {
    'GITHUB_PRIVATE_KEY': PRIVATE_KEY,
}

for key, value in ENV_DEFAULTS.items():
    os.environ.setdefault(key, value)


@pytest.mark.usefixtures('vcrpy_record_mode')
class IGittTestCase(TestCase, metaclass=ABCMeta):
    """
    Base class for writing testcases in IGitt with automated request recording.
    """
    vcr_options = dict()
    enable_unused_record_deletion = True

    @staticmethod
    def remove_link_headers(resp):
        for i, link in enumerate(resp['headers'].get('Link', [])):
            resp['headers']['Link'][i] = FILTER_PARAMS_REGEX.sub(r'\1', link)
        return resp

    @property
    def cassette_name(self):
        """
        Returns the name of the cassette.
        """
        return '{0}.{1}.yaml'.format(self.__class__.__name__,
                                     self._testMethodName)

    @property
    def vcr(self):
        """
        Returns a new vcrpy instance.
        """
        cassettes_dir = join(dirname(getfile(self.__class__)), 'cassettes')
        kwargs = {
            'record_mode': getattr(self, 'vcrpy_record_mode', 'once'),
            'cassette_library_dir': cassettes_dir,
            'match_on': ['method', 'scheme', 'host', 'port', 'path', 'query'],
            'filter_query_parameters': FILTER_QUERY_PARAMS,
            'filter_post_data_parameters': FILTER_QUERY_PARAMS,
            'before_record_response': IGittTestCase.remove_link_headers,
            'filter_headers': ['Link', 'Authorization'],
        }
        kwargs.update(self.vcr_options)
        return VCR(**kwargs)

    @classmethod
    def setUpClass(cls):
        """
        On inherited classes, run `setUp` method as usual.

        Inspired via http://stackoverflow.com/questions/1323455/python-unit-test-with-base-and-sub-class/17696807#17696807
        """
        if cls is not IGittTestCase and cls.setUp is not IGittTestCase.setUp:
            _setup = cls.setUp

            def newSetUp(self, *args, **kwargs):
                IGittTestCase.setUp(self)
                return _setup(self, *args, **kwargs)

            cls.setUp = newSetUp

    def tearDown(self):  # pragma: no cover
        # Check the cassette for unplayed interactions and remove them
        if (not self.cassette.all_played and self.enable_unused_record_deletion
                and not self.cassette.dirty):
            self.cassette.data = [
                v for i, v in enumerate(self.cassette.data)
                if self.cassette.play_counts[i] >= 1
            ]
            self.cassette._save(force=True)

    def setUp(self):
        """
        Common setup method for all inherited classes.
        """
        context_manager = self.vcr.use_cassette(self.cassette_name)
        self.cassette = context_manager.__enter__()
        self.addCleanup(context_manager.__exit__, None, None, None)



"""
Contains the abstraction for a commit in GitHub.
"""
from typing import Optional
from typing import Set

from IGitt import ElementDoesntExistError
from IGitt.GitHub import GitHubMixin, GitHubToken
from IGitt.GitHub.GitHubComment import GitHubComment
from IGitt.GitHub.GitHubRepository import GitHubRepository
from IGitt.GitHub.GitHubIssue import GitHubIssue
from IGitt.Interfaces import get, post
from IGitt.Interfaces.Comment import CommentType
from IGitt.Interfaces.Commit import Commit
from IGitt.Interfaces.CommitStatus import CommitStatus, Status

GH_STATE_TRANSLATION = {Status.ERROR: 'error', Status.FAILED: 'failure',
                        Status.PENDING: 'pending', Status.CANCELED: 'failure',
                        Status.SUCCESS: 'success', Status.RUNNING: 'pending',
                        Status.CREATED: 'pending', Status.MANUAL: 'pending',
                        Status.SKIPPED: 'success'}
INV_GH_STATE_TRANSLATION = {'error': Status.ERROR, 'failure': Status.FAILED,
                            'pending': Status.PENDING,
                            'success': Status.SUCCESS}
GITHUB_KEYWORD_REGEX = {'fix': r'[Ff]ix(?:e[sd])?',
                        'close': r'[Cc]lose[sd]?',
                        'resolve': r'[Rr]esolve[sd]?'}


def get_diff_index(patch, line_nr):
    r"""
    Takes a patch and receives the position of the given line number in it.

    >>> patch = ('---/version/a\n'
    ...          '+++/version/b\n'
    ...          '@@ -1,2 +1,4 @@\n'
    ...          ' # test\n'  # Line 1
    ...          '+\n'  # Line 2
    ...          '-a test repo\n'  # Line 3
    ...          '+something new\n'  # Line 3
    ...          ' something old\n')  # Line 4
    >>> get_diff_index(patch, 1)
    1
    >>> get_diff_index(patch, 3)
    4
    >>> get_diff_index(patch, 4)
    5

    If the line isn't covered in the patch, it'll return None:

    >>> get_diff_index(patch, 8)

    Sometimes patches contain the function and shall still be interpreted
    correctly:

    >>> patch = ('@@ -464,11 +464,10 @@ def get_action_info(section, action):\n'
    ...          ' line 464\n'
    ...          ' line 465\n'
    ...          ' line 466\n'
    ...          '-line 467\n'
    ...          '-line 468\n'
    ...          '+line 467\n'
    ...          '+line 468\n'
    ...          ' line 469\n')
    >>> get_diff_index(patch, 465)
    2
    >>> get_diff_index(patch, 467)
    6

    :param patch: A list of lines of a unified diff.
    :param line_nr: The line number to identify.
    :return: The position in the Patch or None
    """
    patch = patch.splitlines(True)
    current_line_added = 0
    current_diff_index = 0
    for line in patch:
        if line.startswith('---') or line.startswith('+++'):
            continue

        if line.startswith('@@'):
            values = line[line.find('-'):line.find(' @@', 3)]
            _, added = tuple(values.split(' '))
            current_line_added = int(added.split(',')[0][1:])
        elif line.startswith('+') or line.startswith(' '):
            if current_line_added == line_nr:
                return current_diff_index

            current_line_added += 1

        current_diff_index += 1

    return None


class GitHubCommit(GitHubMixin, Commit):
    """
    Represents a commit on GitHub.
    """

    def __init__(self, token: GitHubToken, repository: str, sha: str):
        """
        Creates a new github commit object.

        :param token: A GitHubToken object for authentication.
        :param repository: The full repository name.
        :param sha: The commit SHA.
        """
        self._token = token
        self._repository = repository
        self._sha = sha
        self._url = '/repos/' + repository + '/commits/' + sha

    @property
    def message(self) -> str:
        """
        Returns the commit message.

        :return: Commit message as string.
        """
        return self.data['commit']['message']

    @property
    def sha(self):
        """
        Retrieves the SHA of the commit:

        >>> from os import environ
        >>> commit = GitHubCommit(GitHubToken(environ['GITHUB_TEST_TOKEN']),
        ...                       'gitmate-test-user/test', '674498')
        >>> commit.sha
        '674498'

        :return: A string holding the SHA of the commit.
        """
        return self._sha

    @property
    def repository(self):
        """
        Retrieves the repository that holds this commit.

        >>> from os import environ
        >>> commit = GitHubCommit(GithubToken(environ['GITHUB_TEST_TOKEN']),
        ...                       'gitmate-test-user/test', '3fc4b86')
        >>> commit.repository.full_name
        'gitmate-test-user/test'

        :return: A usable Repository instance.
        """
        return GitHubRepository(self._token, self._repository)

    @property
    def parent(self):
        """
        Retrieves the parent commit. In case of a merge commit the first parent
        will be returned.

        >>> from os import environ
        >>> commit = GitHubCommit(GithubToken(environ['GITHUB_TEST_TOKEN']),
        ...                       'gitmate-test-user/test', '3fc4b86')
        >>> commit.parent.sha
        '674498fd415cfadc35c5eb28b8951e800f357c6f'

        :return: A Commit object.
        """
        return GitHubCommit(self._token, self._repository,
                            self.data['parents'][0]['sha'])

    def set_status(self, status: CommitStatus):
        """
        Adds the given status to the commit.

        >>> from os import environ
        >>> commit = GitHubCommit(GithubToken(environ['GITHUB_TEST_TOKEN']),
        ...                       'gitmate-test-user/test', '3fc4b86')
        >>> status = CommitStatus(Status.FAILED, 'Theres a problem',
        ...                       'gitmate/test')
        >>> commit.set_status(status)
        >>> commit.get_statuses().pop().description
        'Theres a problem'

        If a status with the same context already exists, it will be bluntly
        overridden:

        >>> status.status = Status.SUCCESS
        >>> status.description = "Theres no problem"
        >>> commit.set_status(status)
        >>> len(commit.get_statuses())
        2
        >>> commit.get_statuses().pop().description
        'This commit needs work.'

        :param status: The CommitStatus to set to this commit.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        data = {'state': GH_STATE_TRANSLATION[status.status],
                'target_url': status.url, 'description': status.description,
                'context': status.context}
        status_url = '/repos/' + self._repository + '/statuses/' + self.sha
        post(self._token, self.absolute_url(status_url), data)

    def get_statuses(self) -> Set[CommitStatus]:
        """
        Retrieves the all commit statuses.

        :return: A (frozen)set of CommitStatus objects.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        url = self.url + '/statuses'
        statuses = get(self._token, url)

        # Only the first of each context is the one we want
        result = set()
        contexts = set()
        for status in statuses:
            if status['context'] not in contexts:
                result.add(CommitStatus(
                    INV_GH_STATE_TRANSLATION[status['state']],
                    status['description'], status['context'],
                    status['target_url']))
                contexts.add(status['context'])

        return result

    @property
    def combined_status(self) -> Status:
        """
        Retrieves a combined status of all the commits.

        :return:
            Status.FAILED if any of the commits report as error or failure or
            canceled
            Status.PENDING if there are no statuses or a commit is pending or a
            test is running
            Status.SUCCESS if the latest status for all commits is success
        """
        url = self.url + '/status'
        return INV_GH_STATE_TRANSLATION[get(self._token, url)['state']]

    def get_patch_for_file(self, filename: str):
        r"""
        Retrieves the patch for the given file:

        >>> from os import environ
        >>> commit = GitHubCommit(GithubToken(environ['GITHUB_TEST_TOKEN']),
        ...                       'gitmate-test-user/test', '3fc4b86')
        >>> commit.get_patch_for_file('README.md')
        '@@ -1,2 +1,4 @@\n # test\n a test repo\n+\n+a tst pr'

        But only if it exists!

        >>> commit.get_patch_for_file('isnt there!')
        Traceback (most recent call last):
         ...
        IGitt.ElementDoesntExistError: The file does not exist.

        :param filename: The file to receive the patch for.
        :return: A string containing the patch.
        :raises ElementDoesntExistError: If the given filename doesn't exist.
        """
        for file in self.data['files']:
            if file['filename'] == filename and 'patch' in file:
                return file['patch']

        raise ElementDoesntExistError('The file does not exist.')

    def comment(self, message: str, file: Optional[str]=None,
                line: Optional[int]=None,
                mr_number: Optional[int]=None) -> GitHubComment:
        """
        Places a comment on the commit.

        >>> from os import environ
        >>> commit = GitHubCommit(GithubToken(environ['GITHUB_TEST_TOKEN']),
        ...                       'gitmate-test-user/test', '3fc4b86')

        So this line places a comment on the bottom of the commit,
        not associated to any particular piece of code:

        >>> commit.comment("An issue is here!")

        However, we can also comment on a particular file and line, if that is
        included in the diff:

        >>> commit.comment("Here in line 4, there's a spelling mistake!",
        ...                'README.md', 4)

        If you supply the ``pr_number`` argument, the comment will appear in the
        review UI of that pull request:

        >>> commit.comment("Here in line 4, there's a spelling mistake!",
        ...                'README.md', 4, mr_number=6)

        Beat that! Of course, there's a lot of error handling. If you give the
        wrong file, the comment will appear below the commit with a note about
        the commit, file and line:

        >>> commit.comment("Oh, this'll end up below!!", 'READMENOT.md', 4)

        Also if the line isn't contained in the diff GitHub won't accept that
        and it'll also end up below - sorry!

        >>> commit.comment("Oh, this'll too end up below!!", 'README.md', 8)

        If you give a pull request, the comment will appear on the PR instead:

        >>> commit.comment("Oh, this'll too end up on the PR.",
        ...                'README.md', 8, mr_number=6)

        :param message: The body of the comment.
        :param file: The file to place the comment, relative to repository root.
        :param line: The line in the file in the comment or None.
        :param mr_number: The number of a merge request if this should end up in
                          the review UI of the merge request.
        """
        data = {'body': message}

        if file is not None and line is not None:
            try:
                patch = self.get_patch_for_file(file)
                index = get_diff_index(patch, line)
                if index:  # Else, fallback to comment below file
                    data['position'] = index
                    data['path'] = file
            except ElementDoesntExistError:
                pass  # Fallback to comment below the file

        if 'position' not in data:
            file_str = '' if file is None else ', file ' + file
            line_str = '' if line is None else ', line ' + str(line)
            data['body'] = ('Comment on ' + self.sha + file_str + line_str +
                            '.\n\n' + data['body'])

        comment_type = None

        if mr_number is None:
            comment_type = CommentType.COMMIT
            res = post(self._token, self.url + '/comments', data)
        elif 'position' in data:
            comment_type = CommentType.REVIEW
            data['commit_id'] = self.sha
            res = post(
                self._token,
                self.absolute_url(
                    '/repos/' + self._repository + '/pulls/' + str(mr_number) +
                    '/comments'),
                data)
        else:  # Position not available, pr number available, comment on PR
            comment_type = CommentType.ISSUE
            res = post(
                self._token,
                self.absolute_url(
                    '/repos/' + self._repository + '/issues/' + str(mr_number) +
                    '/comments'),
                data)

        return GitHubComment.from_data(res, self._token, self._repository,
                                       comment_type, res['id'])

    @property
    def unified_diff(self):
        """
        Retrieves the unified diff for the commit excluding the diff index.
        """
        difflines = str(get(self._token, self.url, headers={
            'Accept': 'application/vnd.github.v3.diff'
        })).strip().splitlines()
        # getting rid of the indexing stuff from git diff e.g. removing lines
        # ``diff --git a/somefile b/somefile
        # index 1da2df..2dacdf 100644``
        return '\n'.join([diff for diff in difflines
                          if not diff.startswith('diff --git') and
                          not diff.startswith('index')])

    @property
    def closes_issues(self) -> Set[GitHubIssue]:
        """
        Returns a set of GitHubIssue objects which would be closed upon merging
        this commit.
        """
        return {GitHubIssue(self._token, repo_name, number)
                for number, repo_name in self._get_closes_issues()}

    @property
    def mentioned_issues(self) -> Set[GitHubIssue]:
        """
        Returns a set of GitHubIssue objects which are related to the commit.
        """
        return {GitHubIssue(self._token, repo_name, number)
                for number, repo_name in self._get_mentioned_issues()}


    @property
    def will_fix_issues(self) -> Set[GitHubIssue]:
        """
        Returns a set of GitHubIssue objects which would be fixed as stated in
        this commit message.
        """
        issues =  self.get_keywords_issues(GITHUB_KEYWORD_REGEX['fix'],
                                           [self.message])
        return {GitHubIssue(self._token, repo_name, number)
                for number, repo_name in issues}

    @property
    def will_close_issues(self) -> Set[GitHubIssue]:
        """
        Returns a set of GitHubIssue objects which would be closed as stated in
        this commit message.
        """
        issues =  self.get_keywords_issues(GITHUB_KEYWORD_REGEX['close'],
                                           [self.message])
        return {GitHubIssue(self._token, repo_name, number)
                for number, repo_name in issues}

    @property
    def will_resolve_issues(self) -> Set[GitHubIssue]:
        """
        Returns a set of GitHubIssue objects which would be resolved as stated
        in this commit message.
        """
        issues =  self.get_keywords_issues(GITHUB_KEYWORD_REGEX['resolve'],
                                           [self.message])
        return {GitHubIssue(self._token, repo_name, number)
                for number, repo_name in issues}



"""
This module contains the milestone abstraction class which provides properties
and actions related to milestones.
"""
from datetime import datetime
from IGitt.Interfaces.Repository import Repository
from IGitt.Interfaces import IGittObject
from IGitt.Interfaces import MilestoneStates


class Milestone(IGittObject):
    """
    Represents a milestone for GitHub or GitLab or any similar collection of
    issues.
    """
    @property
    def number(self) -> int:
        """
        Returns the milestone "number" or id.
        """
        raise NotImplementedError

    @property
    def project(self) -> Repository:
        """
        Returns the repository this milestone is linked with.
        """
        raise NotImplementedError

    @property
    def title(self) -> str:
        """
        Retrieves the title of the milestone.
        """
        raise NotImplementedError

    @title.setter
    def title(self, new_title):
        """
        Sets the title of the milestone.

        :param new_title: The new title.
        """
        raise NotImplementedError

    @property
    def description(self) -> str:
        """
        Retrieves the main description of the milestone.
        """
        raise NotImplementedError

    @description.setter
    def description(self, new_description):
        """
        Sets the description of the milestone

        :param new_description: The new description .
        """
        raise NotImplementedError

    @property
    def state(self) -> MilestoneStates:
        """
        Get's the state of the milestone.

        :return: Either MilestoneStates.OPEN or MilestoneStates.CLOSED.
        """
        raise NotImplementedError

    def close(self):
        """
        Closes the milestone.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def reopen(self):
        """
        Reopens the milestone.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    @property
    def created(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone was created.
        """
        raise NotImplementedError

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone was updated the last time.
        """
        raise NotImplementedError

    @property
    def start_date(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone was started.
        """
        raise NotImplementedError

    @start_date.setter
    def start_date(self, new_date: datetime):
        """
        Sets the start date of the milestone.

        :param new_date: The new start date.
        """
        raise NotImplementedError

    @property
    def due_date(self) -> datetime:
        """
        Retrieves a timestamp on when the milestone is due.
        """
        raise NotImplementedError

    @due_date.setter
    def due_date(self, new_date: datetime):
        """
        Sets the due date of the milestone.

        :param new_date: The new due date.
        """
        raise NotImplementedError

    def delete(self):
        """
        Deletes the milestone.
        This is not possible with GitLab api v4.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    @staticmethod
    def create(token, repository, title, body=''):
        """
        Create a new milestone in repository.
        """
        raise NotImplementedError

    @property
    def issues(self) -> set:
        """
        Retrieves a set of issue objects that are assigned to this milestone.
        """
        raise NotImplementedError

    @property
    def merge_requests(self) -> set:
        """
        Retrieves a set of merge request objects that are assigned to this
        milestone.
        """
        raise NotImplementedError



"""
Represents a comment (or note) on GitLab.
"""
from typing import Union
from urllib.parse import quote_plus

from datetime import datetime
from IGitt.GitLab import GitLabMixin
from IGitt.GitLab import GitLabOAuthToken, GitLabPrivateToken
from IGitt.GitLab.GitLabUser import GitLabUser
from IGitt.Interfaces import delete, put
from IGitt.Interfaces.Comment import Comment
from IGitt.Interfaces.Comment import CommentType


class GitLabComment(GitLabMixin, Comment):
    """
    Represents a comment (or note as GitLab folks call it), with mainly a body
    and an author, which can ofcourse be deleted.
    """

    def __init__(self, token: Union[GitLabOAuthToken, GitLabPrivateToken],
                 repository: str, iid: str, comment_type: CommentType,
                 comment_id: int):
        """
        Creates a new GitLabComment with the given data.

        :param token: A Token object to be used for authentication.
        :param repository: The full namespace of the repository.
        :param iid: The unique identifier that links the holder of comment to
                    GitLab. i.e. which identifies the MR (number) or
                    issue (number) or snippet (number) or commit (sha) the
                    comment links to.
        :param comment_type: The type of comment it links to, either one of
                             GitLabComment types.
        :param comment_id: The id of comment.
        """
        self._token = token
        self._repository = repository
        self._type = comment_type
        self._id = comment_id
        self._iid = str(iid)
        self._url = '/projects/{repo}/{c_type}/{iid}/notes/{c_id}'.format(
            repo=quote_plus(repository), c_type=self._type.value,
            iid=iid, c_id=comment_id)
        if comment_type == CommentType.REVIEW:
            raise NotImplementedError

    @property
    def iid(self) -> str:
        """
        Retrieves the internal identifier of the linked resource
        (issue/merge request/commit).
        """
        return self._iid

    @property
    def number(self) -> int:
        """
        Retrieves the id of the comment.
        """
        return self._id

    @property
    def type(self) -> CommentType:
        """
        Retrieves the type of comment it links to.
        """
        return self._type

    @property
    def author(self) -> GitLabUser:
        """
        Retrieves the author of the comment.

        >>> from os import environ
        >>> note = GitLabComment(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                      'gitmate-test-user/test', 1,
        ...                      CommentType.ISSUE, 31500135)
        >>> note.author.username
        'gitmate-test-user'

        :return: A GitLabUser object.
        """
        return GitLabUser.from_data(self.data['author'],
                                    self._token,
                                    self.data['author']['id'])

    @property
    def body(self) -> str:
        r"""
        Retrieves the content of the comment.

        >>> from os import environ
        >>> note = GitLabComment(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                      'gitmate-test-user/test', 1,
        ...                      CommentType.ISSUE, 31500135)
        >>> note.body
        'Lemme comment on you.\r\n'

        :return: A string containing the body.
        """
        return self.data['body']

    @body.setter
    def body(self, value: str):
        """
        Edits the body of the comment on GitLab.

        :param value: A string containing comment body.
        """
        payload = {'body': value}
        self.data = put(self._token, self.url, payload)

    @property
    def created(self) -> datetime:
        """
        Retrieves a timestamp on when the comment was created.

        >>> from os import environ
        >>> note = GitLabComment(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                      'gitmate-test-user/test', 1,
        ...                      CommentType.ISSUE, 31500135)
        >>> note.created
        datetime.datetime(2017, 6, 5, 5, 20, 28, 418000)
        """
        return datetime.strptime(self.data['created_at'],
                                 '%Y-%m-%dT%H:%M:%S.%fZ')

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the comment was updated the last time.

        >>> from os import environ
        >>> note = GitLabComment(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                      'gitmate-test-user/test', 1,
        ...                      CommentType.ISSUE, 31500135)
        >>> note.updated
        datetime.datetime(2017, 6, 5, 6, 5, 34, 491000)
        """
        return datetime.strptime(self.data['updated_at'],
                                 '%Y-%m-%dT%H:%M:%S.%fZ')

    def delete(self):
        """
        Deletes the comment.
        """
        delete(self._token, self.url)

    @property
    def repository(self):
        """
        Returns the GitLab repository this comment was posted in, as a
        GitLabRepository instance.
        """
        from IGitt.GitLab.GitLabRepository import GitLabRepository
        return GitLabRepository(self._token, self._repository)



"""
This package contains the GitLab implementations of the interfaces in
server.git.Interfaces. GitLab drops the support of API version 3 as of
August 22, 2017. So, IGitt adopts v4 to stay future proof.
"""
import os
import logging

from requests_oauthlib import OAuth2

from IGitt.Interfaces import Token, get
from IGitt.Utils import CachedDataMixin


GL_INSTANCE_URL = os.environ.get('GL_INSTANCE_URL', 'https://gitlab.com')
if not GL_INSTANCE_URL.startswith('http'):  # dont cover cause it'll be removed
    GL_INSTANCE_URL = 'https://' + GL_INSTANCE_URL
    logging.warning('Include the protocol in GL_INSTANCE_URL! Omitting it has '
                    'been deprecated.')

BASE_URL = GL_INSTANCE_URL + '/api/v4'


class GitLabMixin(CachedDataMixin):
    """
    Base object for things that are on GitLab.
    """

    def _get_data(self):
        return get(self._token, self.url)

    @staticmethod
    def absolute_url(url):
        """
        Makes a URL like ``/repo/coala/coala`` absolute.
        """
        return BASE_URL + url

    @property
    def hoster(self):
        """
        Tells you that this is a `gitlab` object.
        """
        return 'gitlab'

    @property
    def url(self):
        """
        Returns gitlab API url.
        """
        return self.absolute_url(self._url)

    @property
    def web_url(self):
        """
        Returns a web link for GitLab.
        """
        return self.data['web_url']

    def __repr__(self): # dont cover
        return '<{} object(url={}) at {}>'.format(self.__class__.__name__,
                                                  self.url,
                                                  hex(id(self)))


class GitLabOAuthToken(Token):
    """
    Object representation of OAuth2 tokens.
    """

    def __init__(self, token):
        self._token = token

    @property
    def parameter(self):
        """
        No additional query parameters are used with the token.
        """
        return {}

    @property
    def value(self):
        return self._token

    @property
    def headers(self):
        """
        GitLab OAuth token does not require any special headers.
        """
        return {}

    @property
    def auth(self):
        return OAuth2(token={'access_token': self._token,
                             'token_type': 'bearer'})


class GitLabPrivateToken(Token):
    """
    Object representation of private tokens.
    """

    def __init__(self, token):
        self._token = token

    @property
    def parameter(self):
        return {'private_token': self._token}

    @property
    def value(self):
        return self._token

    @property
    def headers(self):
        """
        GitLab Private token does not require any special headers.
        """
        return {}

    @property
    def auth(self):
        """
        Private token based authorization for GitLab cannot be used directly
        via an AuthBase object.
        """
        return None



"""
Contains a JIRA project implementation.
"""
from datetime import datetime
from urllib.parse import urljoin
from typing import Optional
from typing import Set
from typing import Union
import logging
import re

from IGitt.Interfaces import delete
from IGitt.Interfaces import get
from IGitt.Interfaces import post
from IGitt.Interfaces import AccessLevel
from IGitt.Interfaces import BasicAuthorizationToken
from IGitt.Interfaces import IssueStates
from IGitt.Interfaces.Repository import Repository
from IGitt.Interfaces.Repository import WebhookEvents
from IGitt.Jira import JiraMixin
from IGitt.Jira import JiraOAuth1Token
from IGitt.Jira import BASE_URL
from IGitt.Jira import JIRA_INSTANCE_URL
from IGitt.Jira.JiraIssue import JiraIssue


_HTTP_REGEX = re.compile(r'https?://')
JIRA_WEBHOOK_TRANSLATION = {
    WebhookEvents.PUSH: (None, ),
    WebhookEvents.ISSUE: ('jira:issue_created',
                          'jira:issue_updated',
                          'jira:issue_deleted',
                          'jira:worklog_updated'),
    WebhookEvents.MERGE_REQUEST: (None, ),
    WebhookEvents.COMMIT_COMMENT: (None, ),
    WebhookEvents.MERGE_REQUEST_COMMENT: (None, ),
    WebhookEvents.ISSUE_COMMENT: ('jira:comment_created',
                                  'jira:comment_updated',
                                  'jira:comment_deleted'),
}


class JiraRepository(JiraMixin, Repository):
    """
    Represents a project resource on JIRA.
    """
    def __init__(self,
                 token: Union[JiraOAuth1Token, BasicAuthorizationToken],
                 identifier: Union[str, int]):
        """
        Instantiates a new JiraRepository object with the given details.

        :param token:
            The OAuth v1.0 token to be used for authentication.
        :param identifier:
            The unique identifier or key of the project.
        """
        self._identifier = identifier
        self._token = token
        self._url = '/project/{}'.format(identifier)
        self._hook_url = urljoin(JIRA_INSTANCE_URL,
                                 '/rest/webhooks/1.0/webhook')

    @property
    def identifier(self) -> int:
        """
        Returns the unique identifier of the corresponding repository.
        """
        try:
            return int(self._identifier)
        except ValueError:
            return int(self.data['id'])

    @property
    def web_url(self) -> str:
        """
        Returns a human accessible web url for the corresponding project.
        """
        return urljoin(JIRA_INSTANCE_URL,
                       '/projects/{}'.format(self.data['key']))

    @property
    def top_level_org(self):
        """
        JIRA doesn't support organizations.
        """
        raise NotImplementedError

    @property
    def full_name(self):
        """
        JIRA doesn't have an account based naming scheme like GitHub. So, this
        property returns just the project name.
        """
        return self.data['name']

    @property
    def commits(self):
        """
        JIRA projects don't actually hold a codebase and hence has no commits.
        """
        raise NotImplementedError

    @property
    def clone_url(self):
        """
        JIRA projects don't actually hold a codebase which can be cloned.
        """
        raise NotImplementedError

    def get_labels(self):
        """
        Retrieves the list of available labels for this project.
        """
        raise NotImplementedError

    def create_label(self,
                     name: str,
                     description: Optional[str]=None,
                     label_type: Optional[str]=None,
                     **kwargs):
        """
        Creates a new label.
        """
        raise NotImplementedError

    def delete_label(self, name: str):
        """
        Deletes the specified label.

        :raises: ElementDoesntExistError if no such label exists.
        """
        raise NotImplementedError

    def get_issue(self, issue_number: int):
        """
        Retrieves the specified issue from the project.

        :return:    A JiraIssue object representing the specified issue.
        """
        return JiraIssue(self._token, issue_number)

    def get_mr(self, mr_number: int):
        """
        JIRA doesn't actually hold a codebase and doesn't support sending merge
        requests.
        """
        raise NotImplementedError

    @property
    def hooks(self):
        """
        Retrieves all the webhooks this project is hooked to.

        :return:    A set of URLs.
        """
        params = {'jqlFilter': 'project = {}'.format(self.data['key'])}
        return {hook['url']
                for hook in get(self._token, self._hook_url, params)}

    def register_hook(self,
                      url: str,
                      _: Optional[str]=None,
                      events: Optional[Set[WebhookEvents]]=None):
        """
        Registers a webhook to the given URL.
        """
        if url in self.hooks:
            return

        reg_events = [x for event in events
                      for x in JIRA_WEBHOOK_TRANSLATION[event]
                      if x is not None]
        config = {
            'name': 'An IGitt webhook',
            'url': url,
            'events': reg_events,
            'jqlFilter': 'project = {}'.format(self.data['key']),
            'excludeIssueDetails': False
        }

        post(self._token, self._hook_url, config)
        logging.warning('Be careful with JIRA webhooks as they are system '
                        'wide. i.e. Every single project sends a webhook for '
                        'one registration.')

    def delete_hook(self, url: str):
        """
        Deletes all the webhooks to the specified URL.
        """
        for hook in get(self._token, self._hook_url):
            if hook['url'] == url:
                delete(self._token, hook['self'])

    def filter_issues(self, state: str='opened') -> set:
        """
        Filters issues from a repository based on the chosen properties.
        """
        raise NotImplementedError

    @property
    def issues(self) -> set:
        """
        Retrieves the set of JiraIssue objects for this project.
        """
        params = {'jql': 'Project="{}"'.format(self.data['key'])}
        return {JiraIssue.from_data(iss, self._token, iss['id'])
                for iss in get(self._token,
                               self.absolute_url('/search'),
                               params)['issues']
               }

    @property
    def merge_requests(self) -> set:
        """
        JIRA projects don't actually hold a codebase, so merge requests do not
        exist.
        """
        raise NotImplementedError

    def create_issue(self,
                     issue_type: str,
                     title: str,
                     body: str='') -> JiraIssue:
        """
        Creates a new issue in the project.
        """
        return JiraIssue.create(
            self._token, self.identifier, title, body, issue_type=issue_type)

    def create_fork(self, *args, **kwargs):
        """
        JIRA projects don't actually have a codebase to be forked.
        """
        raise NotImplementedError

    def delete(self):
        """
        Deletes the JIRA project.
        """
        delete(self._token, self.url)

    def create_merge_request(self, *args, **kwargs):
        """
        JIRA projects don't actually hold a codebase to merge changes into.
        """
        raise NotImplementedError

    def create_file(self, *args, **kwargs):
        """
        JIRA projects don't actually hold a codebase. So, new files can't be
        created.
        """
        raise NotImplementedError

    def search_mrs(self, *args, **kwargs):
        """
        JIRA doesn't support a merge request resource.
        """
        raise NotImplementedError

    def search_issues(self,
                      created_after: Optional[datetime]=None,
                      created_before: Optional[datetime]=None,
                      updated_after: Optional[datetime]=None,
                      updated_before: Optional[datetime]=None,
                      state: Optional[IssueStates] = None):
        """
        List open issues in the repository.
        """
        raise NotImplementedError

    def get_permission_level(self, *args, **kwargs) -> AccessLevel:
        """
        JIRA provides support for custom permission schemes for users which
        cannot be therefore generalized into an ``AccessLevel`` enum.
        """
        raise NotImplementedError

    @property
    def parent(self):
        """
        JIRA projects do not support or maintain a fork relationship between
        each other.
        """
        return None

    @staticmethod
    def create(token: Union[JiraOAuth1Token, BasicAuthorizationToken],
               name: str,
               project_type: Optional[str]=None,
               project_key: Optional[str]=None,
               project_lead: Optional[str]=None,
               **kwargs):
        """
        Creates a new JIRA project and returns it.

        :param name:
            The name of the project.
        :param project_type:
            The type of the project.
        :param project_key:
            The key identifier of the project to be set.
        :param project_lead:
            The project lead.
        """
        repo = post(token,
                    BASE_URL + '/project',
                    {
                        'key': project_key,
                        'name': name,
                        'projectTypeKey': project_type,
                        'lead': project_lead
                    })
        return JiraRepository.from_data(repo, token, repo['id'])



"""
This contains the Content implementation for GitHub.
"""
from base64 import b64encode
from typing import Optional

from IGitt.GitHub import GitHubMixin, GitHubToken
from IGitt.Interfaces import get, delete, put
from IGitt.Interfaces.Content import Content


class GitHubContent(GitHubMixin, Content):
    """
    This class represents a content on GitHub
    """
    def __init__(self,  token: GitHubToken, repository: str, path: str):
        self._token = token
        self._repository = repository
        self._url = '/repos/' + repository + '/contents/' + path

    def get_content(self, ref='master'):
        data = {
            'path': self._url,
            'ref': ref
        }
        self.data = get(token=self._token, url=self.url, params=data)

    def delete(self, message: str, branch: Optional[str]=None):
        """
        Deletes content

        :param message: The commit message for the deletion commit.
        :param branch:  The branch to delete this content from. Defaults to
                        `master`.
        """
        if branch is None:
            branch = 'master'

        data = {
            'path': self._url,
            'message' : message,
            'sha': self.data['sha'],
            'branch': branch
        }
        delete(token=self._token, url=self.url, data=data)

    def update(self, message: str, content: str, branch: Optional[str]=None):

        content = b64encode(content.encode()).decode('utf-8')

        if branch is None:
            branch = 'master'

        data = {
            'path': self._url,
            'message': message,
            'sha': self.data['sha'],
            'branch': branch,
            'content' : content
        }
        put(token=self._token, url=self.url, data=data)



"""
This package contains abstractions to use git, as well as services providing
hosting for it (GitHub, GitLab and others).
"""

from os.path import dirname, join


class ElementDoesntExistError(Exception):
    """
    Indicates that the desired element doesn't exist.
    """


class ElementAlreadyExistsError(Exception):
    """
    Indicates that the element (that is probably to be created) already exists.
    """


with open(join(dirname(__file__), 'VERSION'), 'r') as ver:
    VERSION = ver.readline().strip()



"""
Contains the Hoster implementation for GitLab.
"""

from typing import List, Union
import logging

from IGitt.GitLab import GitLabOAuthToken, GitLabPrivateToken, GitLabMixin
from IGitt.GitLab.GitLabComment import GitLabComment
from IGitt.GitLab.GitLabCommit import GitLabCommit
from IGitt.GitLab.GitLabIssue import GitLabIssue
from IGitt.GitLab.GitLabMergeRequest import GitLabMergeRequest
from IGitt.GitLab.GitLabRepository import GitLabRepository
from IGitt.GitLab.GitLabUser import GitLabUser
from IGitt.Interfaces import get
from IGitt.Interfaces import AccessLevel
from IGitt.Interfaces.Actions import IssueActions, MergeRequestActions, \
    PipelineActions
from IGitt.Interfaces.Comment import CommentType
from IGitt.Interfaces.Hoster import Hoster

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


class GitLab(GitLabMixin, Hoster):
    """
    A high level interface to GitLab.
    """

    def __init__(self, token: Union[GitLabOAuthToken, GitLabPrivateToken]):
        """
        Creates a new GitLab Hoster object.

        :param token: A Token object to be used for authentication.
        """
        self._token = token
        self._url = '/'

    @staticmethod
    def _get_repos_with_permissions(repo_list: List[GitLabRepository],
                                    permission: AccessLevel):
        """
        Retrieves repositories the user has permissions to, even inherit the
        permissions for sub-groups and projects.
        """
        # namespaces with permission or greater access level
        namespaces = set()

        # normalize access_levels
        for repo in repo_list:
            if not repo['permissions']['project_access']:
                repo['permissions']['project_access'] = {'access_level': 0}
            if not repo['permissions']['group_access']:
                repo['permissions']['group_access'] = {'access_level': 0}

        # groups with access_level > permission
        to_check = [proj['namespace']['id'] for proj in repo_list if
                    proj['permissions']['group_access'].get(
                        'access_level', 0) >= permission.value]

        # finding subgroups
        for namespace in to_check:
            namespaces.add(namespace)
            # add sub_groups of this namespace
            to_check += [proj['namespace']['id'] for proj in repo_list if
                         proj['namespace']['parent_id'] == namespace]

        return [repo for repo in repo_list
                if repo['namespace']['id'] in namespaces or
                repo['permissions']['project_access'].get(
                    'access_level', 0) >= permission.value]

    @property
    def master_repositories(self):
        """
        Retrieves repositories the user has admin access to.
        """
        repo_list = get(self._token, self.absolute_url('/projects'),
                        {'membership': True})
        return {GitLabRepository.from_data(repo, self._token,
                                           repo['path_with_namespace'])
                for repo in
                self._get_repos_with_permissions(repo_list,
                                                 AccessLevel.ADMIN)}

    @property
    def owned_repositories(self):
        """
        Retrieves repositories owned by the authenticated user.

        >>> from os import environ
        >>> GitLab = GitLab(GitLabOAuthToken(viron['GITLAB_TEST_TOKEN']))
        >>> sorted(map(lambda x: x.full_name, GitLab.owned_repositories)
        {'gitmate-test-user/test'}

        :return: A set of GitLabRepository objects.
        """
        repo_list = get(self._token, self.absolute_url('/projects'),
                        {'owned': True})
        return {GitLabRepository.from_data(repo, self._token,
                                           repo['path_with_namespace'])
                for repo in repo_list}

    @property
    def write_repositories(self):
        """
        Retrieves the full names of repositories this user can write to.

        >>> from os import environ
        >>> GitLab = GitLab(GitLabOAuthToken(viron['GITLAB_TEST_TOKEN']))
        >>> sorted(map(lambda x: x.full_name, GitLab.write_repositories))
        ['gitmate-test-user/test', 'nkprince007/gitmate-test']

        :return: A set of GitLabRepository objects.
        """
        repo_list = get(self._token, self.absolute_url('/projects'),
                        {'membership': True})
        return {GitLabRepository.from_data(repo, self._token,
                                           repo['path_with_namespace'])
                for repo in
                self._get_repos_with_permissions(
                    repo_list,
                    AccessLevel.CAN_WRITE)}

    def get_repo(self, repository) -> GitLabRepository:
        """
        Retrieve a given repository.

        >>> from os import environ
        >>> GitLab = GitLab(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']))
        >>> repo = source.get_repo('gitmate-test-user/test')
        >>> isinstance(repo, GitLabRepository)
        True

        :return: A repository object.
        """
        return GitLabRepository(self._token, repository)

    def get_user(self, username: str=None) -> GitLabUser:
        """
        Retrieves repository name from given webhook data.

        :param username:    Username to fetch, use None to fetch authenticated
                            user.
        :return:            The user.
        """
        return GitLabUser(self._token, identifier=username)

    @staticmethod
    def get_repo_name(webhook: dict):
        """
        Retrieves the repository name from given webhook data.
        """
        # Push, Tag, Issue, Note, Wiki Page and Pipeline Hooks
        if 'project' in webhook.keys():
            return webhook['project']['path_with_namespace']

        # Merge Request Hook
        if 'object_attributes' in webhook.keys():
            return webhook['object_attributes']['target']['path_with_namespace']

        # Build Hook
        if 'repository' in webhook.keys():
            ssh_url = webhook['repository']['git_ssh_url']
            return ssh_url[ssh_url.find(':') + 1: ssh_url.rfind('.git')]

    @staticmethod
    def raw_search(token: Union[GitLabPrivateToken, GitLabOAuthToken],
                   raw_query: str):
        """
        GitLab doesn't allow searching through raw queries.
        """
        raise NotImplementedError


    @staticmethod
    def _handle_labels(actions_enum: Union[IssueActions, MergeRequestActions],
                       obj_to_return: Union[GitLabIssue, GitLabMergeRequest],
                       data: dict):
        """
        Yields `LABELED` or `UNLABELED` actions for each label added or removed
        from given `Issue` or `MergeRequest`.
        """
        old_attrs = {label['title']
                     for label in
                     data['changes']['labels']['previous']}
        new_attrs = {label['title']
                     for label in
                     data['changes']['labels']['current']}

        # new labels added
        for label in new_attrs - old_attrs:
            yield actions_enum.LABELED, [obj_to_return, label]

        # labels removed
        for label in old_attrs - new_attrs:
            yield actions_enum.UNLABELED, [obj_to_return, label]


    def _handle_webhook_issue(self, data, repository):
        issue = data['object_attributes']
        issue_obj = GitLabIssue.from_data(
            issue,
            self._token, repository, issue['iid'])
        trigger_event = {
            'open': IssueActions.OPENED,
            'close': IssueActions.CLOSED,
            'reopen': IssueActions.REOPENED,
        }.get(issue['action'], IssueActions.ATTRIBUTES_CHANGED)

        if (trigger_event == IssueActions.ATTRIBUTES_CHANGED and
                'labels' in data['changes']):
            # labels are changed
            yield from type(self)._handle_labels(
                IssueActions, issue_obj, data)
        else:
            yield trigger_event, [issue_obj]

    def _handle_webhook_merge_request(self, data, repository):
        merge_request_data = data['object_attributes']
        merge_request_obj = GitLabMergeRequest.from_data(
            merge_request_data,
            self._token,
            repository,
            merge_request_data['iid'])
        trigger_event = {
            'update': MergeRequestActions.ATTRIBUTES_CHANGED,
            'open': MergeRequestActions.OPENED,
            'reopen': MergeRequestActions.REOPENED,
            'merge': MergeRequestActions.MERGED,
            'close': MergeRequestActions.CLOSED,
        }.get(merge_request_data['action'])

        # nasty workaround for finding merge request resync
        if 'oldrev' in merge_request_data:
            trigger_event = MergeRequestActions.SYNCHRONIZED


        # no such webhook event action implemented yet
        if not trigger_event:
            raise NotImplementedError('Unrecgonized action: Merge Request Hook'
                                      '/' + merge_request_data['action'])

        if (trigger_event is MergeRequestActions.ATTRIBUTES_CHANGED and
                'labels' in data['changes']):
            yield from type(self)._handle_labels(MergeRequestActions,
                                                 merge_request_obj, data)
        else:
            yield trigger_event, [merge_request_obj]

    def _handle_webhook_note(self, data, repository):
        comment = data['object_attributes']
        comment_type = {
            'MergeRequest': CommentType.MERGE_REQUEST,
            'Commit': CommentType.COMMIT,
            'Issue': CommentType.ISSUE,
            'Snippet': CommentType.SNIPPET
        }.get(comment['noteable_type'])

        if comment_type == CommentType.MERGE_REQUEST:
            iid = data['merge_request']['iid']
            iss = GitLabMergeRequest.from_data(data['merge_request'],
                                               self._token, repository, iid)
            action = MergeRequestActions.COMMENTED
        elif comment_type == CommentType.ISSUE:
            iid = data['issue']['iid']
            iss = GitLabIssue.from_data(data['issue'], self._token,
                                        repository, iid)
            action = IssueActions.COMMENTED
        else:
            raise NotImplementedError

        yield action, [iss, GitLabComment.from_data(
            comment,
            self._token, repository, iid, comment_type, comment['id']
        )]

    def _handle_webhook_pipeline(self, data, repository):
        yield PipelineActions.UPDATED, [GitLabCommit(
            self._token,
            repository,
            data['commit']['id'])]

    def handle_webhook(self, event: str, data: dict):
        """
        Handles a GitLab webhook for you.

        If it's an issue event it returns e.g.
        ``IssueActions.OPENED, [GitLabIssue(...)]``, for comments it returns
        ``MergeRequestActions.COMMENTED,
        [GitLabMergeRequest(...), GitLabComment(...)]``, for updates it returns
        ``IssueActions.LABELED, [GitLabIssue(...), 'new label']``

        :param event:       The HTTP_X_GITLAB_EVENT of the request header.
        :param data:        The pythonified JSON data of the request.
        :yields:            An IssueActions or MergeRequestActions member and a
                            list of the affected IGitt objects.
        """
        repository = self.get_repo_name(data)
        part_handler_name = '_'.join(
            event.strip('Hook').strip().lower().split())

        try:
            handler = getattr(self, '_handle_webhook_' +  part_handler_name)
        except AttributeError:
            raise NotImplementedError('Given webhook cannot be handled yet.')
        else:
            yield from handler(data, repository)



"""
This contains the Issue implementation for GitLab.
"""
from datetime import datetime
from datetime import timedelta
from typing import List
from typing import Optional
from typing import Set
from typing import Union
from urllib.parse import quote_plus

from IGitt.GitLab import GitLabMixin
from IGitt.GitLab import GitLabOAuthToken, GitLabPrivateToken
from IGitt.GitLab.GitLabComment import GitLabComment
from IGitt.GitLab.GitLabReaction import GitLabReaction
from IGitt.GitLab.GitLabUser import GitLabUser
from IGitt.Interfaces.Comment import CommentType
from IGitt.Interfaces.Issue import Issue
from IGitt.Interfaces import get, put, post, delete
from IGitt.Interfaces import IssueStates
from IGitt.Interfaces import MergeRequestStates


class GitLabIssue(GitLabMixin, Issue):
    """
    This class represents an issue on GitLab.
    """

    def __init__(self, token: Union[GitLabOAuthToken, GitLabPrivateToken],
                 repository: str, number: int):
        """
        Creates a new GitLabIssue with the given credentials.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/repo_that_doesnt_exist', 1)
        ... # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        RuntimeError: ({'message': 'Not Found', ...}, 404)

        :param token: A Token object to be used for authentication.
        :param repository: The full name of the repository.
                           e.g. ``sils/baritone``.
        :param number: The issue internal identification number.
        :raises RuntimeError: If something goes wrong (network, auth, ...)
        """
        self._token = token
        self._repository = repository
        self._iid = number
        self._url = '/projects/{repo}/issues/{issue_iid}'.format(
            repo=quote_plus(repository), issue_iid=number)

    @property
    def repository(self):
        """
        Returns the GitLab repository this issue is linked with as a
        GitLabRepository instance.
        """
        from IGitt.GitLab.GitLabRepository import GitLabRepository
        return GitLabRepository(self._token, self._repository)

    @property
    def title(self) -> str:
        """
        Retrieves the title of the issue.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)

        >>> issue.title
        'Take it serious, son!'

        You can simply set it using the property setter:

        >>> issue.title = 'dont panic'
        >>> issue.title
        'dont panic'

        >>> issue.title = 'Take it serious, son!'

        :return: The title of the issue - as string.
        """
        return self.data['title']

    @title.setter
    def title(self, new_title):
        """
        Sets the title of the issue.

        :param new_title: The new title.
        """
        self.data = put(self._token, self.url, {'title': new_title})

    @property
    def number(self) -> int:
        """
        Returns the issue "number" or id.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> issue.number
        1

        :return: The number of the issue.
        """
        return self._iid

    @property
    def assignees(self):
        """
        Retrieves the assignee of the issue:

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> issue.assignees
        {'gitmate-test-user'}

        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 2)
        >>> issue.assignees # Returns empty set, unassigned
        {}

        :return: A set containing the usernames of assignees.
        """
        return {GitLabUser.from_data(user, self._token, user['id'])
                for user in self.data['assignees']}

    def assign(self, *usernames: List[GitLabUser]):
        """
        Adds the user as one of the assignees of the issue.
        :param users: User objects of the users to be added as an assignee.
        """
        self.assignees |= set(usernames)

    def unassign(self, *users: List[GitLabUser]):
        """
        Removes the user from the assignees of the issue.
        :param users: User objects of the users to be unassigned.
        """
        self.assignees = self.assignees - set(users)

    @assignees.setter
    def assignees(self, value: Set[GitLabUser]):
        """
        Setter for assignees.
        """
        self.data = put(self._token, self.url,
                        {'assignee_ids': [u.identifier for u in value]})

    @property
    def available_assignees(self) -> Set[GitLabUser]:
        """
        Retrieves a set of available assignees for the issue.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> {a.username for a in issue.available_assignees}
        {'gitmate-test-user'}

        :return: A set of GitLabUsers.
        """
        return {
            GitLabUser.from_data(user, self._token, user['id'])
            for user in get(
                self._token, self.absolute_url(
                    '/projects/' + quote_plus(self._repository) + '/members'))
        }

    @property
    def description(self) -> str:
        r"""
        Retrieves the main description of the issue.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> issue.description
        'I am a serious issue. Fix me soon, dude.'

        :return: A string containing the main description of the issue.
        """
        return self.data['description'] if self.data['description'] else ''

    @description.setter
    def description(self, new_description):
        """
        Sets the description of the issue.

        :param new_description: The new description.
        """
        self.data = put(self._token, self.url, {'description': new_description})

    @property
    def author(self) -> GitLabUser:
        """
        Retrieves the author of the issue.

        :return: A GitLabUser object.
        """
        return GitLabUser.from_data(self.data['author'],
                                    self._token,
                                    self.data['author']['id'])

    def add_comment(self, body):
        """
        Adds a comment to the issue.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> comment = issue.add_comment("Doh!")

        You can use the comment right after:

        >>> comment.body
        'Doh!'
        >>> comment.delete()

        The comment will be created by the user authenticated via the oauth
        token.

        :param body: The body of the new comment to create.
        :return: The newly created comment.
        """
        result = post(self._token, self.url + '/notes', {'body': body})

        return GitLabComment(self._token, self._repository, self.number,
                             CommentType.ISSUE, result['id'])

    @property
    def comments(self) -> List[GitLabComment]:
        r"""
        Retrieves comments from the issue.

        As of now, the list of comments is not sorted. Related issue on GitLab
        CE here - https://gitlab.com/gitlab-org/gitlab-ce/issues/32057

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 3)
        >>> comments = issue.comments
        >>> for comment in comments:
        ...     print(comment.body)
        Stop staring at me.
        Go, get your work done.

        :return: A list of Comment objects.
        """
        return [
            GitLabComment.from_data(
                result, self._token, self._repository, self.number,
                CommentType.ISSUE, result['id']
            )
            for result in get(self._token, self.url + '/notes')
        ]

    @property
    def labels(self) -> Set[str]:
        """
        Retrieves all labels associated with this bug.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> issue.labels
        set()

        Use the property setter to set the labels:

        >>> issue.labels = {'a', 'b', 'c'}
        >>> sorted(issue.labels)
        ['a', 'b', 'c']

        Use the empty set intuitively to clear all labels:

        >>> issue.labels = set()

        :return: A list of label captions (str).
        """
        return set(self.data['labels'])

    @labels.setter
    def labels(self, value: Set[str]):
        """
        Sets the value of labels to the given set of labels.

        :param value: A set of label texts.
        """
        # Only if self.data is populated we actually save a request here
        if 'labels' in self.data and value == self.labels:
            return  # No need to patch

        self.data = put(self._token, self.url,
                        {'labels': ','.join(map(str, value))})

    @property
    def available_labels(self) -> Set[str]:
        """
        Retrieves a set of captions that are available for labelling bugs.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> sorted(issue.available_labels)
        ['a', 'b', 'c']

        :return: A set of label captions (str).
        """
        return {label['name'] for label in get(
            self._token, self.absolute_url(
                '/projects/' + quote_plus(self._repository) + '/labels'))}

    @property
    def created(self)->datetime:
        """
        Retrieves a timestamp on when the issue was created.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 4)
        >>> issue.created
        datetime.datetime(2017, 6, 5, 9, 45, 20, 678000)
        """
        return datetime.strptime(self.data['created_at'],
                                 '%Y-%m-%dT%H:%M:%S.%fZ')

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the issue was updated the last time.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 4)
        >>> issue.updated
        datetime.datetime(2017, 6, 5, 9, 45, 56, 115000)
        """
        return datetime.strptime(self.data['updated_at'],
                                 '%Y-%m-%dT%H:%M:%S.%fZ')

    def close(self):
        """
        Closes the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        self.data = put(self._token, self.url, {'state_event': 'close'})

    def reopen(self):
        """
        Reopens the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        self.data = put(self._token, self.url, {'state_event': 'reopen'})

    def delete(self):
        """
        Deletes the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        delete(self._token, self.url)

    @property
    def state(self):
        """
        Get's the state of the issue.

        >>> from os import environ
        >>> issue = GitLabIssue(GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...                     'gitmate-test-user/test', 1)
        >>> issue.state
        <IssueStates.OPEN: 'open'>
        >>> str(issue.state)
        'open'

        So if we close it:

        >>> issue.close()
        >>> issue.state
        <IssueStates.CLOSED: 'closed'>
        >>> str(issue.state)
        'closed'

        And reopen it:

        >>> issue.reopen()
        >>> issue.state
        <IssueStates.OPEN: 'open'>


        Note: GitLab Issues & Merge Requests API underwent a change to have
        only two states, <IssueStates.OPEN: 'open'> or
        <IssueStates.CLOSED: 'closed'>. No 'reopened' state anymore.

        :return:
            Either <IssueStates.OPEN: 'open'> or
            <IssueStates.CLOSED: 'closed'>.
        """
        if self.data['state'] == 'opened':
            self.data['state'] = 'open'

        return IssueStates[self.data['state'].upper()]

    @property
    def reactions(self) -> Set[GitLabReaction]:
        """
        Retrieves the reactions / award emojis applied on the issue.
        """
        url = self.url + '/award_emoji'
        reactions = get(self._token, url)
        return {GitLabReaction.from_data(r, self._token, self, r['id'])
                for r in reactions}

    @staticmethod
    def create(token: Union[GitLabOAuthToken, GitLabPrivateToken],
               repository: str,
               title: str,
               body: str='',
               issue_type: Optional[str]=None):
        """
        Create a new issue with given title and body.

        >>> from os import environ
        >>> issue = GitLabIssue.create(
        ...     GitLabOAuthToken(environ['GITLAB_TEST_TOKEN']),
        ...     'gitmate-test-user/test',
        ...     'test issue title',
        ...     'sample description'
        ... )
        >>> issue.state
        <IssueStates.OPEN: 'open'>

        Delete the issue to avoid filling the test repo with issues.

        >>> issue.delete()

        :return: GitLabIssue object of the newly created issue.
        """
        url = '/projects/{repo}/issues'.format(repo=quote_plus(repository))
        issue = post(token, GitLabIssue.absolute_url(url),
                     {'title': title, 'description': body})

        return GitLabIssue.from_data(issue, token, repository, issue['iid'])

    @property
    def mrs_closed_by(self):
        """
        Returns the merge requests that close this issue.
        """
        from IGitt.GitLab.GitLabMergeRequest import GitLabMergeRequest

        url = '{url}/closed_by'.format(url=self._url)
        mrs = get(self._token, self.absolute_url(url))

        return {GitLabMergeRequest.from_data(mr,
                                             self._token,
                                             self._repository,
                                             mr['iid'])
                for mr in mrs if mr['state'] == MergeRequestStates.MERGED.value}

    @property
    def milestone(self):
        """
        Retrieves the milestone.
        """
        from IGitt.GitLab.GitLabProjectMilestone import GitLabProjectMilestone
        return GitLabProjectMilestone.from_data(
            self.data['milestone'], self._token, self._repository,
            self.data['milestone']['id']
        ) if self.data['milestone'] else None

    @milestone.setter
    def milestone(self, new_milestone):
        """
        Setter for the Milestone.
        Delete the Milestone with passing a 'None'
        """
        self.data = put(
            self._token, self.url,
            {'milestone_id': new_milestone.number if new_milestone else ''})

    @property
    def time_estimate(self) -> timedelta:
        """
        Retrieves the time_estimate from the 'time_estimate' attribute
        in the GitLab Issue.
        """
        return timedelta(seconds=self.data['time_stats']['time_estimate'])

    @time_estimate.setter
    def time_estimate(self, new_time_estimate: timedelta):
        """
        Setter for the time_estimate.
        Sets the time on 'human_time_estimate' since it's not possible to set
        'time_estimate'. GitLab then automatically calculates 'time_estimate'.
        time_estimate is set to 0 when passing a None or 0.
        """
        if new_time_estimate:
            human_time_estimate = \
                            str(int(new_time_estimate.total_seconds())) + 's'
            self.data = post(self._token, self.url +
                             '/time_estimate',
                             {'duration': human_time_estimate})
        else:
            self.data = post(self._token, self.url +
                             '/reset_time_estimate', {'duration': None})

    @property
    def total_time_spent(self) -> timedelta:
        """
        Retrieves the time_estimate from the 'total_time_spent' attribute
        in the GitLab Issue.
        """
        return timedelta(seconds=self.data['time_stats']['total_time_spent'])

    @total_time_spent.setter
    def total_time_spent(self, absolute_time_spent: timedelta):
        """
        Writes the value of absolute_time_spent into total_time_spent.
        Can't be less than 0.
        Allows any time unit of the timedelta object.
        Allows total_time_spent to be reset to 0 by passing a None or 0.
        """
        if absolute_time_spent and absolute_time_spent.total_seconds() != 0:
            self.data = post(self._token, self.url +
                             '/reset_spent_time', {'duration': None})
            human_time_spent = \
                        str(int(absolute_time_spent.total_seconds())) + 's'
            self.data = post(self._token, self.url +
                             '/add_spent_time',
                             {'duration': human_time_spent})
        else:
            self.data = post(self._token, self.url +
                             '/reset_spent_time', {'duration': None})

    def add_to_total_time_spent(self, relative_time_spent: timedelta):
        """
        Adds the value of relative_time_spent to total_time_spent.
        Allows for positive and negative values.
        Allows any time unit of the timedelta object.
        Does nothing when passing None or 0.
        """
        if relative_time_spent and relative_time_spent.total_seconds() != 0:
            human_time_spent = \
                        str(int(relative_time_spent.total_seconds())) + 's'
            self.data = post(self._token, self.url +
                             '/add_spent_time',
                             {'duration': human_time_spent})



"""
Contains the Repository class.
"""

from datetime import datetime
from enum import Enum
from os import chdir, getcwd
from tempfile import mkdtemp
from typing import List
from typing import Optional
from typing import Set
from typing import Union

from git.repo.base import Repo

from IGitt.Interfaces import AccessLevel
from IGitt.Interfaces import IGittObject
from IGitt.Interfaces import MergeRequestStates
from IGitt.Interfaces import IssueStates
from IGitt.Interfaces import Token


class WebhookEvents(Enum):
    """
    This class depicts the webhook events that can be registered with any
    hosting service providers like GitHub or GitLab.
    """
    PUSH = 1
    ISSUE = 2
    MERGE_REQUEST = 3
    ISSUE_COMMENT = 4
    COMMIT_COMMENT = 5
    MERGE_REQUEST_COMMENT = 6


class Repository(IGittObject):
    """
    This class depicts a Repository at a hosting service like GitHub. Thus, on
    top of access to the actual code and history, it also provides access to
    issues, PRs, hooks and so on.
    """

    @property
    def identifier(self) -> int:
        """
        Returns the identifier of the repository.
        """
        raise NotImplementedError

    @property
    def top_level_org(self):
        """
        Returns the topmost organization, e.g. for `gitmate/open-source/IGitt`
        this is `gitmate`.
        """
        raise NotImplementedError

    def register_hook(self,
                      url: str,
                      secret: Optional[str]=None,
                      events: Optional[Set[WebhookEvents]]=None):
        """
        Registers a webhook to the given URL. Should pass silently if the hook
        already exists.

        :param url: The URL to fire the webhook to.
        :param secret: An optional secret token to be registered with webhook.
        :param events: The set of events for which the webhook is to be
                       registered against. Defaults to all possible events.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def delete_hook(self, url: str):
        """
        Deletes all webhooks to the given URL. (Does nothing if no such hook
        exists.)

        :param url: The URL to not fire the webhook to anymore.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    @property
    def hooks(self) -> Set[str]:
        """
        Retrieves all URLs this repository is hooked to.

        :return: Set of URLs (str).
        """
        raise NotImplementedError

    def get_issue(self, issue_number: int):
        """
        Retrieves an issue.

        :param issue_number: The issue ID of the issue to retrieve.
        :return: An Issue object.
        :raises ElementDoesntExistError: If the issue doesn't exist.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def get_mr(self, mr_number: int):
        """
        Retrieves an MR.

        :param mr_number: The merge_request ID of the MR to retrieve.
        :return: An MR object.
        """
        raise NotImplementedError

    def create_label(self,
                     name: str,
                     color: Optional[str]=None,
                     description: Optional[str]=None,
                     label_type: Optional[str]=None):
        """
        Creates a new label.

        :param name:
            The name of the label to create.
        :param color:
            A HTML color value with a leading #.
        :param description:
            The description of the label to be created.
        :param label_type:
            The type of the label to be created.
        :raises ElementAlreadyExistsError:
            If the label name already exists.
        :raises RuntimeError:
            If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def delete_label(self, name: str):
        """
        Deletes a label.

        :param name: The caption of the label to delete.
        :raises ElementDoesntExistError: If the label doesn't exist.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def get_clone(self) -> Union[Repo, str]:
        """
        Clones the repository into a temporary directory:

        >>> test_repo = type(
        ...     'MockRepo', (Repository,),
        ...     {'clone_url': 'https://github.com/sils/configurations'})
        >>> repo, path = test_repo().get_clone()

        With this Repo object you can easily access the source code of the
        repository as well as all commits:

        >>> type(repo)
        <class 'git.repo.base.Repo'>
        >>> repo.branches
        [<git.Head "refs/heads/master">]

        Or simply access it via the path if you care only for the head:

        >>> from os.path import exists, join
        >>> assert exists(join(path, '.gitignore'))

        Be sure to not forget to clean it up in the end:

        >>> from shutil import rmtree
        >>> rmtree(path)

        :return: A tuple containing a Repo object and the path to the
                 repository.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        tempdir = mkdtemp()

        # Workaround for
        # https://github.com/gitpython-developers/GitPython/issues/734
        try:
            getcwd()
        except FileNotFoundError:
            chdir(tempdir)

        repo = Repo.clone_from(self.clone_url, tempdir)

        return repo, tempdir

    def get_labels(self) -> Set[str]:
        """
        Retrieves the set of labels.

        :return: A set of strings, the captions of the labels.
        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    @property
    def full_name(self) -> str:
        """
        Retrieves the full name of the repository, that identifies it uniquely
        at the hoster.

        :return: A string, e.g. 'coala-analyzer/coala'.
        """
        raise NotImplementedError

    @property
    def filter_commits(self, author: Optional[str]=None):
        """
        Filter commits based on properties.

        :author: Author of the commit.
        :return: Set of Commit objects.
        """
        raise NotImplementedError

    @property
    def commits(self):
        """
        Retrieves the set of commits in this repository.

        :return: Set of Commit objects.
        """
        raise NotImplementedError

    @property
    def clone_url(self) -> str:
        """
        Retrieves an url that can be used for cloning the repository.

        :return: A string that can be used with ``git clone <url>`` as url.
        """
        raise NotImplementedError

    def filter_merge_requests(self, state: str='opened') -> set:
        """
        Filters the merge requests from the repository based on the state
        of the merge requests.

        :param state: 'merged' or 'opened' or 'closed' or 'all'
        """
        raise NotImplementedError

    @property
    def merge_requests(self) -> set:
        """
        Retrieves a set of open merge request objects.
        """
        raise NotImplementedError

    def filter_issues(self,
                      state: str='opened',
                      label: Optional[str]=None,
                      assignee: Optional[str]=None
                     ) -> set:
        """
        Filters the issues from the repository based on properties.

        :param state: 'opened' or 'closed' or 'all'.
        :param label: Label of the issue
        :param assignee: username of issue assignee
        """
        raise NotImplementedError

    @property
    def issues(self) -> set:
        """
        Retrieves a set of issue objects.
        """
        raise NotImplementedError

    def create_issue(self, title, body=''):
        """
        Create a new issue.
        """
        raise NotImplementedError

    def create_fork(self, organization: Optional[str]=None,
                    namespace: Optional[str]=None):
        """
        Create a fork
        """
        raise NotImplementedError

    def delete(self):
        """
        Delete Repository
        """
        raise NotImplementedError

    def create_merge_request(self, title: str, base: str, head: str,
                             body: Optional[str]=None,
                             target_project_id: Optional[int]=None,
                             target_project: Optional[str]=None):
        """
        Creates a PR
        """
        raise NotImplementedError

    def create_file(self, path: str, message: str, content: str,
                    branch: Optional[str]=None, committer: Optional[str]=None,
                    author: Optional[dict]=None, encoding: Optional[str]=None):
        """
        Creates a new file
        """
        raise NotImplementedError

    def search_mrs(self,
                   created_after: Optional[datetime]=None,
                   created_before: Optional[datetime]=None,
                   updated_after: Optional[datetime]=None,
                   updated_before: Optional[datetime]=None,
                   state: Optional[MergeRequestStates]=None):
        """
        Retrieves a list of prs
        """
        raise NotImplementedError

    def search_issues(self,
                      created_after: Optional[datetime]=None,
                      created_before: Optional[datetime]=None,
                      updated_after: Optional[datetime]=None,
                      updated_before: Optional[datetime]=None,
                      state: Optional[IssueStates]=None):
        """
        Retrieves a list of issues
        """
        raise NotImplementedError

    def get_permission_level(self, user) -> AccessLevel:
        """
        Retrieves the permission level for the specified user on this
        repository.
        """
        raise NotImplementedError

    @property
    def parent(self):
        """
        Returns the repository from which this repository is forked from.
        Returns `None` if it has no fork relationship.
        """
        raise NotImplementedError

    @staticmethod
    def create(token: Token,
               name: str,
               path: Optional[str]=None,
               namespace_id: Optional[int]=None,
               default_branch: Optional[str]='master',
               resolve_outdated_diff_discussions: bool=False,
               import_url: Optional[str]=None,
               public_jobs: bool=False,
               tag_list: Optional[List[str]]=None,
               avatar: Optional[object]=None,
               ci_config_path: Optional[str]=None,
               repository_storage: Optional[str]=None,
               approvals_before_merge: Optional[int]=None,
               org_name: Optional[str]=None,
               description: Optional[str]=None,
               homepage: Optional[str]=None,
               visibility: str='public',
               has_issues: bool=True,
               has_merge_requests: bool=True,
               has_jobs: bool=True,
               has_snippets: bool=True,
               has_projects: bool=True,
               has_wiki: bool=True,
               has_container_registry: bool=True,
               has_shared_runners: bool=True,
               has_lfs: bool=False,
               team_id: Optional[int]=None,
               auto_init: bool=False,
               gitignore_template: Optional[str]=None,
               license_template: Optional[str]=None,
               allow_request_access: bool=True,
               allow_squash_merge: bool=True,
               allow_merge_commit: bool=True,
               allow_rebase_merge: bool=True,
               allow_printing_merge_request_link: bool=True,
               only_allow_merge_if_pipeline_succeeds: bool=False,
               project_type: Optional[str]=None,
               project_key: Optional[str]=None,
               project_lead: Optional[str]=None):
        """
        Creates a new repository and returns it.
        """
        raise NotImplementedError



"""
Represents the comment resource from JIRA.
"""
from datetime import datetime
from typing import Union

from IGitt.Interfaces import delete
from IGitt.Interfaces import put
from IGitt.Interfaces import BasicAuthorizationToken
from IGitt.Interfaces.Comment import Comment
from IGitt.Interfaces.Comment import CommentType
from IGitt.Interfaces.User import User
from IGitt.Jira import JiraMixin
from IGitt.Jira import JiraOAuth1Token


class JiraComment(JiraMixin, Comment):
    """
    Represents a comment on JIRA, mainly with a body and author - oh and it's
    deletable!
    """

    def __init__(self,
                 token: Union[BasicAuthorizationToken, JiraOAuth1Token],
                 issue_id: Union[int, str],
                 comment_id: int):
        """
        Instantiates a new JiraComment instance from the given data.

        :param token:
            The token to be used for authentication.
        :param issue_id:
            The unique identifier or key for the issue on JIRA.
        :param comment_id:
            The unique identifier of the comment on JIRA.
        """
        self._token = token
        self._id = comment_id
        self._issue = issue_id
        self._url = '/issue/{}/comment/{}'.format(self._issue, self._id)

    @property
    def number(self) -> int:
        """
        Retrieves the id of the comment.
        """
        return self._id

    @property
    def type(self) -> CommentType:
        """
        Retrieves the type of the comment it links to.
        """
        return CommentType.ISSUE

    @property
    def body(self):
        """
        Retrieves the content of the comment:

        :return: A string containing the body.
        """
        return self.data['body']

    @body.setter
    def body(self, new_body: str):
        """
        Modifies the comment body to ``new_body``.

        :param new_body: A string containing comment body.
        """
        payload = {'body': new_body}
        self.data = put(self._token, self.url, payload)

    @property
    def author(self) -> User:
        """
        Retrieves the author of the comment.

        :return: A User object.
        """
        raise NotImplementedError

    @property
    def created(self) -> datetime:
        """
        Retrieves a timestamp on when the comment was created.

        :return:    A datetime.datetime object representing the time of
                    creation of the comment.
        """
        return datetime.strptime(self.data['created'],
                                 '%Y-%m-%dT%H:%M:%S.%f%z')

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the comment was updated the last time.

        :return:    A datetime.datetime object representing the time of the
                    last update of the comment.
        """
        return datetime.strptime(self.data['updated'],
                                 '%Y-%m-%dT%H:%M:%S.%f%z')

    def delete(self):
        """
        Deletes the comment.
        """
        delete(self._token, self.url)

    @property
    def repository(self):
        """
        Returns the JIRA project this comment was posted in, as a
        JiraRepository instance.
        """
        raise NotImplementedError



"""
This module contains the Issue abstraction class which provides properties and
actions related to issues and bug reports.
"""
from typing import Set
from typing import Optional

from IGitt.Interfaces import Token
from IGitt.Interfaces import IGittObject
from IGitt.Interfaces.User import User
from IGitt.Interfaces.Issue import Issue
from IGitt.Interfaces.Repository import Repository


class Organization(IGittObject):
    """
    Represents an organization on GitHub or GitLab.
    """
    @property
    def identifier(self) -> int:
        """
        Returns the identifier of the organization.
        """
        raise NotImplementedError

    @property
    def description(self) -> str:
        """
        Returns the description of the Organization.
        """
        raise NotImplementedError

    @property
    def billable_users(self) -> int:
        """
        Number of paying/registered users on the organization.
        """
        raise NotImplementedError

    @property
    def owners(self) -> Set[User]:
        """
        Returns the user handles of all admin users, usually the owner role.
        """
        raise NotImplementedError

    @property
    def masters(self) -> Set[User]:
        """
        Returns the user handles of all users able to manage members, usually
        the master role (sometimes no such role exists, then same as owners).
        """
        raise NotImplementedError

    @property
    def name(self) -> str:
        """
        The name of the organization.
        """
        raise NotImplementedError

    @property
    def suborgs(self):
        """
        Returns the sub-organizations within this repository.
        """
        raise NotImplementedError

    @property
    def repositories(self) -> Set[Repository]:
        """
        Returns the list of repositories contained in this organization.
        """
        raise NotImplementedError

    def filter_issues(self,
                      state: Optional[str]='opened',
                      label: Optional[str]=None,
                      assignee: Optional[str]=None) -> Set[Issue]:
        """
        Filters the issues in the organization based on properties

        :param state: 'opened' or 'closed' or 'all'
        :param label: Label of the issue
        :param assignee: username of issue assignee
        :return: Set of Issue objects
        """
        raise NotImplementedError

    @property
    def issues(self) -> Set[Issue]:
        """
        Returns set of issue objects in this organization.
        """
        raise NotImplementedError

    @staticmethod
    def create(token: Token,
               name: str,
               path: str,
               parent_id: Optional[int]=None,
               description: Optional[str]=None,
               visibility: str='private',
               lfs_enabled: bool=False,
               request_access_enabled: bool=False):
        """
        Creates a new organization from the given parameters.

        :param token:
            The credentials to be used for authorization.
        :param name:
            The name of the organization.
        :param path:
            The path of the organization.
        :param parent_id:
            The parent organization id to create nested organization.
        :param description:
            The description of the organization.
        :param visibility:
            Controls the visibility of the organization. Can be either
            'private', 'public' or 'internal'.
        :param lfs_enabled:
            Enables Git Large File System for projects in this organization.
        :param request_access_enabled:
            Allow users to request member access on the organization.
        """
        raise NotImplementedError



"""
This contains the Notification implementation for GitHub.

Take note that GitHub Notifications are actually available via Threads API and
Notifications API is just a wrapper to fetching these threads.
"""
from typing import Union

from IGitt.GitHub import BASE_URL
from IGitt.GitHub import GitHubMixin
from IGitt.GitHub import GitHubToken
from IGitt.GitHub.GitHubCommit import GitHubCommit
from IGitt.GitHub.GitHubIssue import GitHubIssue
from IGitt.GitHub.GitHubMergeRequest import GitHubMergeRequest
from IGitt.GitHub.GitHubRepository import GitHubRepository
from IGitt.Interfaces import delete
from IGitt.Interfaces import get
from IGitt.Interfaces import patch
from IGitt.Interfaces.Notification import Notification
from IGitt.Interfaces.Notification import Reason


class GitHubNotification(GitHubMixin, Notification):
    """
    This class represents a Notification on GitHub.
    """

    def __init__(self, token: GitHubToken, identifier: Union[str, int]):
        """
        Creates a new GitHubNotification object with the given credentials.
        """
        self._token = token
        self._id = str(identifier)
        self._url = '/notifications/threads/' + self._id

    @property
    def identifier(self) -> str:
        """
        Returns the id of the notification thread from GitHub.
        """
        return self._id

    @property
    def reason(self) -> Reason:
        """
        Returns the reason for notification.
        """
        return {
            'assign': Reason.ASSIGNED,
            'author': Reason.AUTHORED,
            'comment': Reason.COMMENTED,
            'invitation': Reason.INVITED,
            'mention': Reason.MENTIONED,
            'manual': Reason.MANUAL,
            'state_change': Reason.STATE_CHANGED,
            'subscribed': Reason.SUBSCRIBED,
            'team_mention': Reason.MENTIONED
        }[self.data['reason']]

    @property
    def subject_type(self) -> type:
        """
        Returns the type of the subject the notification.
        """
        return {
            'Commit': GitHubCommit,
            'Issue': GitHubIssue,
            'PullRequest': GitHubMergeRequest
        }[self.data['subject']['type']]

    @property
    def subject(self) -> Union[GitHubCommit, GitHubIssue, GitHubMergeRequest]:
        """
        Returns the subject of the notification.

        :return: A GitHubCommit, GitHubIssue or GitHubMergeRequest object.
        """
        subject_data = self.data['subject']
        identifier = subject_data['url'].strip().split('/')[-1]
        return self.subject_type.from_data(subject_data, self._token,
                                           self.repository.full_name,
                                           identifier)

    @property
    def repository(self) -> GitHubRepository:
        """
        Returns the repository this notification is related to.
        """
        return GitHubRepository.from_data(self.data['repository'],
                                          self._token,
                                          self.data['repository']['full_name'])

    @property
    def pending(self) -> bool:
        """
        Returns True if the notification is pending/not read.
        """
        return self.data['unread'] or False

    def unsubscribe(self):
        """
        Unsubscribe from this subject.
        """
        delete(self._token, self.url + '/' + 'subscription')

    def mark_done(self):
        """
        Marks the notification as done/read.
        """
        patch(self._token, self.url, {})
        self.data.update({'unread': False})

    @staticmethod
    def fetch_all(token: GitHubToken):
        """
        Returns the list of notifications for the user bearing the token.
        """
        return [GitHubNotification.from_data(notif, token, notif['id'])
                for notif in get(token, BASE_URL + '/notifications')]



"""
Contains the Team abstraction class.
"""
from typing import Set

from IGitt.Interfaces import IGittObject
from IGitt.Interfaces.User import User


class Team(IGittObject):
    """
    Represents a team on GitHub or GitLab.
    """
    @property
    def name(self) -> str:
        """
        Name of the team.
        """
        raise NotImplementedError

    @property
    def id(self) -> int:
        """
        Retrieves the id of the team.
        """
        raise NotImplementedError

    @property
    def description(self) -> str:
        """
        Returns the description of this team.
        """
        raise NotImplementedError

    @property
    def members(self) -> Set[User]:
        """
        Returns the user handles of all members of this team.
        """
        raise NotImplementedError

    def is_member(self, username: str):
        """
        Checks if given username is member of this team.
        """
        raise NotImplementedError

    @property
    def get_organization(self):
        """
        Returns parent organization.
        """
        raise NotImplementedError



"""
Holds the Reaction class representing reactions on issues, comments and merge
requests.
"""
from IGitt.Interfaces import IGittObject
from IGitt.Interfaces.User import User


class Reaction(IGittObject):
    """
    Represents a reaction / award emoji on GitHub and GitLab.
    """

    @property
    def name(self) -> str:
        """
        Returns the name of the reaction.
        """
        raise NotImplementedError

    @property
    def user(self) -> User:
        """
        Returns the user who reacted with this.
        """
        raise NotImplementedError



#!/usr/bin/env python3

"""
Provides a main routine that can install IGitt and create distributions.
"""
from subprocess import call
import sys

import setuptools.command.build_py
from setuptools import find_packages, setup

from IGitt import VERSION


class BuildDocsCommand(setuptools.command.build_py.build_py):
    """
    Automatically generates the documentation with sphinx.
    """
    apidoc_command = (
        'sphinx-apidoc', '-f', '-o', 'docs', '--no-toc', 'IGitt'
    )
    doc_command = ('make', '-C', 'docs', 'html', 'SPHINXOPTS=-W')

    def run(self):
        errOne = call(self.apidoc_command)
        errTwo = call(self.doc_command)
        sys.exit(errOne or errTwo)


with open('requirements.txt') as requirements:
    REQUIRED = requirements.read().splitlines()

with open('README.rst') as readme:
    long_description = readme.read()


if __name__ == '__main__':
    setup(name='IGitt',
          cmdclass={'docs': BuildDocsCommand},
          version=VERSION,
          description='A git(hub/lab/...) hosting abstraction library.',
          long_description=long_description,
          url='https://gitlab.com/gitmate/open-source/IGitt/',
          author='Lasse Schuirmann',
          maintainer='Lasse Schuirmann',
          maintainer_email='lasse@gitmate.io',
          packages=find_packages(exclude=['build.*', '*.tests.*', '*.tests']),
          install_requires=REQUIRED,
          package_data={'IGitt': ['VERSION']},
          license='MIT')



"""
This contains the Issue implementation for Jira.
"""
from datetime import datetime
from typing import List
from typing import Optional
from typing import Set
from typing import Union
from urllib.parse import urljoin

from IGitt.Interfaces import delete
from IGitt.Interfaces import get
from IGitt.Interfaces import post
from IGitt.Interfaces import put
from IGitt.Interfaces import IssueStates
from IGitt.Interfaces.Issue import Issue
from IGitt.Interfaces.User import User
from IGitt.Jira import JiraMixin
from IGitt.Jira import JiraOAuth1Token
from IGitt.Jira import JIRA_INSTANCE_URL
from IGitt.Jira.JiraComment import JiraComment


class JiraIssue(JiraMixin, Issue):
    """
    This class represents an Issue resource on JIRA.
    """

    def __init__(self, token: JiraOAuth1Token, identifier: Union[str, int]):
        """
        Instantiates a new JiraIssue object from the given details.

        :param token:
            The OAuth v1.0 token to be used for authentication with JIRA.
        :param identifier:
            The unique identifier or key related to the issue resource.
        """
        self._identifier = identifier
        self._token = token
        self._url = '/issue/{}'.format(identifier)

    @property
    def number(self):
        """
        Retrieves the unique identifier of the issue.

        :return: An integer representing the id of the issue.
        """
        try:
            return int(self._identifier)
        except ValueError:
            return int(self.data['id'])

    @property
    def web_url(self):
        """
        Returns a human accessible web url for the corresponding issue.

        :return:    A string containing the web link for the issue.
        """
        return urljoin(JIRA_INSTANCE_URL,
                       '/browse/{}'.format(self.data['key']))

    @property
    def title(self):
        """
        Retrieves the title or summary of the issue.

        :return: A string containing the title of summary of the issue.
        """
        return self.data['fields']['summary']

    @title.setter
    def title(self, new_title: str):
        """
        Sets the title or summary of the issue to the specified title.

        :param new_title:   The new title to be set on the issue.
        """
        data = {'update': {'summary': [{'set': new_title}]}}
        put(self._token, self.url, data)
        self.data['fields']['summary'] = new_title

    @property
    def description(self):
        """
        Retrieves the description of the issue.

        :return: A string containing the description of the issue.
        """
        return self.data['fields']['description']

    @description.setter
    def description(self, new_description):
        """
        Sets the description of the issue.

        :param new_description: The new description.
        """
        data = {'update': {'description': [{'set': new_description}]}}
        put(self._token, self.url, data)
        self.data['fields']['summary'] = new_description

    @property
    def assignees(self):
        """
        Retrieves the assignees of the issue.

        Note: JIRA only allows one assignee per issue.

        :return: A set of usernames of assignees.
        """
        return {self.data['fields']['assignee']['name']}

    @assignees.setter
    def assignees(self, value: Set[User]):
        """
        Setter for ssignees.
        """
        raise NotImplementedError

    def assign(self, *usernames: List[User]):
        """
        Sets a given users as assignee.
        """
        raise NotImplementedError

    def unassign(self, *usernames: List[User]):
        """
        Unassigns given users from issue.
        """
        raise NotImplementedError

    @property
    def labels(self) -> Set[str]:
        """
        Retrieves the set of labels the issue is currently tagged with.

        :return: The set of labels.
        """
        return set(self.data['fields'].get('labels', []))

    @labels.setter
    def labels(self, value: Set[str]):
        """
        Tags the issue with the given labels. For examples see documentation
        of the labels read function.

        Labels are added and removed as necessary on remote.

        :param value: The new set of labels.
        """
        data = {'update': {'labels': [{'set': list(value)}]}}
        put(self._token, self.url, data)
        self.data['fields']['labels'] = value

    @property
    def available_labels(self) -> Set[str]:
        """
        Compiles a set of labels that are available for labelling this issue.

        :return: A set of label captions.
        """
        raise NotImplementedError

    @property
    def author(self):
        """
        Retrieves the author of the issue.

        :return: A string containing the username of author.
        """
        return self.data['fields']['creator']['name']

    @property
    def comments(self) -> Set[JiraComment]:
        """
        Retrieves the comments on the issue.

        :return: A set of JiraComment instances.
        """
        return {
            JiraComment.from_data(comment,
                                  self._token,
                                  self._identifier,
                                  comment['id'])
            for comment in get(self._token, self.url + '/comment')['comments']
        }

    def add_comment(self, body) -> JiraComment:
        """
        Adds a comment to the issue.

        :param body:            The content of the comment.
        :return:                The newly created comment.
        :raises RuntimeError:   If something goes wrong (network, auth...).
        """
        url = self.url + '/comment'
        resp = post(self._token, url, data={'body': body})
        return JiraComment.from_data(
            resp, self._token, self._identifier, resp['id'])

    @property
    def state(self):
        """
        Retrieves the state of the issue. JIRA determines the state of an issue
        based on ``resolution`` property of issue resource. If ``resolution``
        is ``null``, then the issue is still open, otherwise it has been closed
        or resolved.

        :return:    Either IssueStates.OPEN or IssueStates.CLOSED.
        """
        return (IssueStates.CLOSED if self.data['fields']['resolution']
                else IssueStates.OPEN)

    def close(self):
        """
        Closes the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    def reopen(self):
        """
        Reopens the issue.

        :raises RuntimeError: If something goes wrong (network, auth...).
        """
        raise NotImplementedError

    @property
    def created(self) -> datetime:
        """
        Retrieves a timestamp on when the comment was created.

        :return:    A datetime.datetime object representing the time of
                    creation of the issue.
        """
        return datetime.strptime(self.data['fields']['created'],
                                 '%Y-%m-%dT%H:%M:%S.%f%z')

    @property
    def updated(self) -> datetime:
        """
        Retrieves a timestamp on when the comment was updated the last time.

        :return:    A datetime.datetime object representing the last updated
                    time of the issue.
        """
        return datetime.strptime(self.data['fields']['updated'],
                                 '%Y-%m-%dT%H:%M:%S.%f%z')

    @property
    def reactions(self) -> List[str]:
        """
        Retrieves the reactions / award emojis applied on the issue.
        """
        raise NotImplementedError

    @property
    def mrs_closed_by(self) -> Set:
        """
        Returns the merge requests that close this issue.
        """
        raise NotImplementedError

    def delete(self):
        """
        Deletes the issue.
        """
        delete(self._token, self.url)

    @staticmethod
    def create(token: JiraOAuth1Token,
               project: str,
               title: str,
               body: str='',
               issue_type: Optional[str]=None):
        """
        Creates a new issue with the given title and body on the specified
        project.

        :param token:
            The OAuth 1.0 token to be used for authentication.
        :param project:
            The identifier of the project in which the issue should be opened.
        :param issue_type:
            The name of the issue type resource to be used.
        :param title:
            The title of the issue.
        :param body:
            The description of the issue.
        :return:
            A new JiraIssue object.
        """
        if not issue_type:
            raise TypeError('`issue_type` cannot be empty.')
        post_url = '/issue/'
        data = {
            'description': body,
            'fields': {
                'summary': title,
                'issuetype': {'name': issue_type},
                'project': {'id': project}
            }
        }

        resp = post(token, JiraIssue.absolute_url(post_url), data)
        return JiraIssue(token, resp['id'])



"""
Here you go: GitHub organizations can be used in IGitt.
"""
from functools import lru_cache
from typing import Set
from typing import Optional
from urllib.parse import quote_plus

from IGitt.GitHub import GitHubToken
from IGitt.GitHub import GH_INSTANCE_URL
from IGitt.GitHub import GitHubMixin
from IGitt.GitHub.GitHubIssue import GitHubIssue
from IGitt.GitHub.GitHubUser import GitHubUser
from IGitt.Interfaces import get
from IGitt.Interfaces.Organization import Organization
from IGitt.Interfaces.Repository import Repository


GH_ISSUE_STATE_TRANSLATION = {
    'opened': 'open',
    'closed': 'closed',
    'all': 'all'
}


class GitHubOrganization(GitHubMixin, Organization):
    """
    Represents an organization on GitHub.
    """

    @property
    def web_url(self):
        return '{}/{}'.format(GH_INSTANCE_URL, self.name)

    def __init__(self, token, name):
        """
        :param name: Name of the org.
        """
        self._token = token
        self._name = name
        self._url = '/orgs/{name}'.format(name=quote_plus(name))

    @property
    def identifier(self) -> int:
        """
        Returns the identifier of the organization.
        """
        return self.data['id']

    @property
    def description(self) -> str:
        """
        Returns the description of this organization.
        """
        return self.data['description']

    @property
    def billable_users(self) -> int:
        """
        Number of paying/registered users on the organization.
        """
        try:
            return len(get(self._token, self.url + '/members'))
        except RuntimeError:
            return 1

    @property
    def owners(self) -> Set[GitHubUser]:
        """
        Returns the user handles of all admin users.
        """
        try:
            return {
                GitHubUser.from_data(user, self._token, user['login'])
                for user in get(
                    self._token, self.url + '/members',
                    params={'role': 'admin'}
                )
            }
        except RuntimeError:
            return {GitHubUser(self._token, self.name)}

    @property
    def masters(self) -> Set[GitHubUser]:
        """
        Gets all owners (because there's no masters role on GitHub).
        """
        return self.owners

    @property
    def name(self) -> str:
        """
        The name of the organization.
        """
        return self._name

    @property
    def suborgs(self) -> Set[Organization]:
        """
        Returns the sub-organizations within this repository.
        """
        return set()

    @property
    @lru_cache(None)
    def repositories(self) -> Set[Repository]:
        """
        Returns the list of repositories contained in this organization.
        """
        from IGitt.GitHub.GitHubRepository import GitHubRepository

        return {GitHubRepository.from_data(repo, self._token, repo['id'])
                for repo in get(self._token, self.url + '/repos')}

    def filter_issues(self,
                      state: Optional[str]=None,
                      label: Optional[str]=None,
                      assignee: Optional[str]=None) -> Set[GitHubIssue]:
        """
        Filters the issues in the organization based on properties

        :param state: 'opened' or 'closed' or 'all'.
        :param label: Label of the issue
        :param assignee: username of issue assignee
        :return: Set of GitHubIssue objects
        """
        from IGitt.GitHub.GitHub import GitHub
        query = 'user:' + self.name
        if state:
            query += ' state:' + GH_ISSUE_STATE_TRANSLATION[state]
        if label:
            query += ' label:' + label
        if assignee:
            query += ' assignee:' + assignee
        return set(GitHub.raw_search(self._token, query))

    @property
    def issues(self) -> Set[GitHubIssue]:
        """
        Returns the set of Issues in this organization.
        """
        return self.filter_issues(state='opened')

    @staticmethod
    def create(token: GitHubToken,
               name: str,
               path: str,
               parent_id: Optional[int]=None,
               description: Optional[str]=None,
               visibility: str='private',
               lfs_enabled: bool=False,
               request_access_enabled: bool=False) -> Organization:
        """
        GitHub does not allow creating new organizations via the API.
        """
        raise NotImplementedError('GitHub does not allow creating '
                                  'new organizations via the API as of now.')
